import io
import logging
from typing import Optional, TypeVar

from ofrak.component.analyzer import Analyzer
from ofrak.core import NamedProgramSection
from ofrak.core.architecture import ProgramAttributes
from ofrak.core.elf.model import (
    ElfSectionHeader,
    Elf,
    ElfHeader,
    ElfBasicHeader,
    ElfProgramHeader,
    ElfSegmentStructure,
    ElfSegment,
    ElfSectionStructure,
    ElfSection,
    UnanalyzedElfSection,
    ElfSymbol,
    ElfSymbolStructure,
    ElfRelaEntry,
    ElfDynamicEntry,
    ElfVirtualAddress,
    UnanalyzedElfSegment,
)
from ofrak.core.memory_region import MemoryRegion
from ofrak.model.component_model import ComponentConfig
from ofrak.model.resource_model import ResourceAttributes
from ofrak.model.viewable_tag_model import AttributesType
from ofrak.resource import Resource
from ofrak.service.resource_service_i import ResourceFilter
from ofrak_io.deserializer import BinaryDeserializer
from ofrak_type.range import Range

LOGGER = logging.getLogger(__name__)

RA = TypeVar("RA", bound=ResourceAttributes)


class ElfBasicHeaderAttributesAnalyzer(Analyzer[None, ElfBasicHeader]):
    """
    Deserialize the [ElfBasicHeader][ofrak.core.elf.model.ElfBasicHeader], which contains the
    first 7 fields of the ELF header. These fields are all endianness- & word-size-agnostic,
    and in fact define the endianness and word size for the remainder of the header.

    The remaining fields are deserialized as part of the
    [ElfHeader][ofrak.core.elf.model.ElfHeader]. See "ELF header (Ehdr)" in
    <https://man7.org/linux/man-pages/man5/elf.5.html> for details.
    """

    id = b"ElfHeaderMetadataAttributesAnalyzer"
    targets = (ElfBasicHeader,)
    outputs = (ElfBasicHeader,)

    async def analyze(self, resource: Resource, config=None) -> ElfBasicHeader:
        tmp = await resource.get_data()
        deserializer = BinaryDeserializer(io.BytesIO(tmp))
        (
            ei_magic,
            ei_class,
            ei_data,
            ei_version,
            ei_osabi,
            ei_abiversion,
            ei_pad,
        ) = deserializer.unpack_multiple("4sBBBBB7s")
        assert ei_magic == b"\x7fELF"
        return ElfBasicHeader(
            ei_magic, ei_class, ei_data, ei_version, ei_osabi, ei_abiversion, ei_pad
        )


class ElfHeaderAttributesAnalyzer(Analyzer[None, ElfHeader]):
    """
    Deserialize the [ElfHeader][ofrak.core.elf.model.ElfHeader], which contains all of the
    ELF header fields except the first 7. The first 7 fields are deserialized as part of the
    [ElfBasicHeader][ofrak.core.elf.model.ElfBasicHeader].
    """

    id = b"ElfHeaderAttributesAnalyzer"
    targets = (ElfHeader,)
    outputs = (ElfHeader,)

    async def analyze(self, resource: Resource, config=None) -> ElfHeader:
        deserializer = await _create_deserializer(resource)
        return self.deserialize(deserializer)

    @classmethod
    def deserialize(cls, deserializer: BinaryDeserializer) -> ElfHeader:
        (
            e_type,
            e_machine,
            e_version,
            e_entry,
            e_phoff,
            e_shoff,
            e_flags,
            e_ehsize,
            e_phentsize,
            e_phnum,
            e_shentsize,
            e_shnum,
            e_shstrndx,
        ) = deserializer.unpack_multiple(f"HHIQQQIHHHHHH", auto_bitwidth=True)

        return ElfHeader(
            e_type,
            e_machine,
            e_version,
            e_entry,
            e_phoff,
            e_shoff,
            e_flags,
            e_ehsize,
            e_phentsize,
            e_phnum,
            e_shentsize,
            e_shnum,
            e_shstrndx,
        )


class ElfProgramHeaderAttributesAnalyzer(Analyzer[None, ElfProgramHeader]):
    """
    Deserialize an [ElfProgramHeader][ofrak.core.elf.model.ElfProgramHeader].
    """

    id = b"ElfProgramHeaderAttributesAnalyzer"
    targets = (ElfProgramHeader,)
    outputs = (ElfProgramHeader,)

    async def analyze(self, resource: Resource, config=None) -> ElfProgramHeader:
        segment_structure = await resource.view_as(ElfSegmentStructure)
        deserializer = await _create_deserializer(resource)
        return self.deserialize(deserializer, segment_structure.segment_index)

    @classmethod
    def deserialize(cls, deserializer: BinaryDeserializer, elf_index: int) -> ElfProgramHeader:
        p_flags = 0
        p_type = deserializer.unpack_uint()
        if deserializer.get_word_size() == 8:
            p_flags = deserializer.unpack_uint()
        p_offset, p_vaddr, p_paddr, p_filesz, p_memsz = deserializer.unpack_multiple(
            "QQQQQ", auto_bitwidth=True
        )
        if deserializer.get_word_size() == 4:
            p_flags = deserializer.unpack_uint()
        p_align = deserializer.unpack_ulong()
        return ElfProgramHeader(
            elf_index, p_type, p_offset, p_vaddr, p_paddr, p_filesz, p_memsz, p_flags, p_align
        )


class ElfSegmentAnalyzer(Analyzer[None, ElfSegment]):
    """
    Analyze an ElfSegment.
    """

    id = b"ElfSegmentAnalyzer"
    targets = (ElfSegment,)
    outputs = (ElfSegment,)

    async def analyze(self, resource: Resource, config=None) -> ElfSegment:
        unnamed_segment = await resource.view_as(UnanalyzedElfSegment)
        segment_header = await unnamed_segment.get_header()
        return ElfSegment(
            segment_index=segment_header.segment_index,
            virtual_address=segment_header.p_vaddr,
            size=segment_header.p_memsz,
        )


class ElfSectionHeaderAttributesAnalyzer(Analyzer[None, ElfSectionHeader]):
    """
    Deserialize an [ElfSectionHeader][ofrak.core.elf.model.ElfSectionHeader].
    """

    id = b"ElfSectionHeaderAttributesAnalyzer"
    targets = (ElfSectionHeader,)
    outputs = (ElfSectionHeader,)

    async def analyze(self, resource: Resource, config=None) -> ElfSectionHeader:
        section_structure = await resource.view_as(ElfSectionStructure)
        deserializer = await _create_deserializer(resource)
        return self.deserialize(deserializer, section_structure.section_index)

    @classmethod
    def deserialize(cls, deserializer: BinaryDeserializer, elf_index: int) -> ElfSectionHeader:
        (
            sh_name,
            sh_type,
            sh_flags,
            sh_addr,
            sh_offset,
            sh_size,
            sh_link,
            sh_info,
            sh_addralign,
            sh_entsize,
        ) = deserializer.unpack_multiple("IIQQQQIIQQ", auto_bitwidth=True)
        return ElfSectionHeader(
            elf_index,
            sh_name,
            sh_type,
            sh_flags,
            sh_addr,
            sh_offset,
            sh_size,
            sh_link,
            sh_info,
            sh_addralign,
            sh_entsize,
        )


class ElfSymbolAttributesAnalyzer(Analyzer[None, ElfSymbol]):
    """
    Deserialize an [ElfSymbol][ofrak.core.elf.model.ElfSymbol], an entry in the ELF symbol
    table.
    """

    id = b"ElfSymbolAnalyzer"
    targets = (ElfSymbol,)
    outputs = (ElfSymbol,)

    async def analyze(self, resource: Resource, config=None) -> ElfSymbol:
        symbol_structure = await resource.view_as(ElfSymbolStructure)
        deserializer = await _create_deserializer(resource)

        return self.deserialize(deserializer, symbol_structure.symbol_index)

    @classmethod
    def deserialize(cls, deserializer: BinaryDeserializer, elf_index: int) -> ElfSymbol:
        if deserializer.get_word_size() == 8:
            st_name, st_info, st_other, st_shndx, st_value, st_size = deserializer.unpack_multiple(
                "IBBHQQ"
            )
        else:
            st_name, st_value, st_size, st_info, st_other, st_shndx = deserializer.unpack_multiple(
                "IIIBBH"
            )
        return ElfSymbol(elf_index, st_name, st_value, st_size, st_info, st_other, st_shndx)


class ElfDynamicSectionAnalyzer(Analyzer[None, ElfDynamicEntry]):
    """
    If an object file participates in dynamic linking, its program header table will have an element
    of type PT_DYNAMIC. This segment contains the .dynamic section.

    Descriptions of the Dynamic Table entries may be found herein:
    https://docs.oracle.com/cd/E19683-01/817-3677/chapter6-42444/index.html
    """

    id = b"ElfDynamicSectionAnalyzer"
    targets = (ElfDynamicEntry,)
    outputs = (ElfDynamicEntry,)

    async def analyze(self, resource: Resource, config=None) -> ElfDynamicEntry:
        deserializer = await _create_deserializer(resource)
        return self.deserialize(deserializer)

    @classmethod
    def deserialize(cls, deserializer: BinaryDeserializer) -> ElfDynamicEntry:
        if deserializer.get_word_size() == 8:
            d_tag, d_un = deserializer.unpack_multiple("QQ")
        else:
            d_tag, d_un = deserializer.unpack_multiple("II")
        return ElfDynamicEntry(d_tag, d_un)


class ElfPointerAnalyzer(Analyzer[None, ElfVirtualAddress]):
    id = b"ElfPointerAnalyzer"
    targets = (ElfVirtualAddress,)
    outputs = (ElfVirtualAddress,)

    async def analyze(self, resource: Resource, config=None) -> ElfVirtualAddress:
        deserializer = await _create_deserializer(resource)
        return self.deserialize(deserializer)

    @classmethod
    def deserialize(cls, deserializer: BinaryDeserializer) -> ElfVirtualAddress:
        value = deserializer.unpack_ulong()
        return ElfVirtualAddress(value)


class ElfRelaAnalyzer(Analyzer[None, ElfRelaEntry]):
    """
    Deserialize an [ElfRelaEntry][ofrak.core.elf.model.ElfRelaEntry], an entry in a rela.*
    table.

    http://sourceware.org/git/?p=glibc.git;a=blob_plain;f=elf/elf.h
    """

    id = b"ElfRelaAnalyzer"
    targets = (ElfRelaEntry,)
    outputs = (ElfRelaEntry,)

    async def analyze(self, resource: Resource, config=None) -> ElfRelaEntry:
        deserializer = await _create_deserializer(resource)
        return self.deserialize(deserializer)

    @classmethod
    def deserialize(cls, deserializer: BinaryDeserializer) -> ElfRelaEntry:
        if deserializer.get_word_size() == 8:
            r_offset, r_info, r_addend = deserializer.unpack_multiple("QQq")
        else:
            r_offset, r_info, r_addend = deserializer.unpack_multiple("IIi")
        return ElfRelaEntry(r_offset, r_info, r_addend)


class ElfSectionNameAnalyzer(Analyzer[None, AttributesType[NamedProgramSection]]):
    """
    Get the name of an ELF section. ELF section names are stored as null-terminated strings in
    dedicated string section, and each ELF section header's `sh_name` field is an offset in this
    section.
    """

    id = b"ElfSectionNameAnalyzer"
    targets = (ElfSection,)
    outputs = (AttributesType[NamedProgramSection],)

    async def analyze(self, resource: Resource, config=None) -> AttributesType[NamedProgramSection]:
        unnamed_section = await resource.view_as(UnanalyzedElfSection)
        section_header = await unnamed_section.get_header()
        elf_r = await unnamed_section.get_elf()
        string_section = await elf_r.get_section_name_string_section()
        try:
            string_section_data = await string_section.resource.get_data(
                Range(section_header.sh_name, Range.MAX)
            )
            name_string_end = string_section_data.find(b"\x00")
            raw_section_name = await string_section.resource.get_data(
                Range(section_header.sh_name, section_header.sh_name + name_string_end)
            )
            section_name = raw_section_name.decode("ascii")
        except ValueError:
            LOGGER.info("String section is empty! Using '<no-strings>' as section name")
            section_name = "<no-strings>"  # This is what readelf returns in this situation
        return AttributesType[NamedProgramSection](
            name=section_name,
        )


class ElfSectionMemoryRegionAnalyzer(Analyzer[None, MemoryRegion]):
    """
    Get the in-memory address and size of an ELF section. These are stored in the corresponding
    ELF section header.
    """

    id = b"ElfSectionMemoryRegionAnalyzer"
    targets = (ElfSection,)
    outputs = (MemoryRegion,)

    async def analyze(self, resource: Resource, config=None) -> MemoryRegion:
        unnamed_section = await resource.view_as(UnanalyzedElfSection)
        section_header = await unnamed_section.get_header()
        return MemoryRegion(
            virtual_address=section_header.sh_addr,
            size=section_header.sh_size,
        )


class ElfProgramAttributesAnalyzer(Analyzer[None, ProgramAttributes]):
    """
    Analyze the `ProgramAttributes` of an ELF, which are part of the information stored in the ELF
    header.
    """

    id = b"ElfProgramAttributesAnalyzer"
    targets = (Elf,)
    outputs = (ProgramAttributes,)

    async def analyze(
        self, resource: Resource, config: Optional[ComponentConfig] = None
    ) -> ProgramAttributes:
        elf_header = await resource.get_only_descendant_as_view(
            ElfHeader, r_filter=ResourceFilter.with_tags(ElfHeader)
        )
        elf_basic_header = await resource.get_only_descendant_as_view(
            ElfBasicHeader, r_filter=ResourceFilter.with_tags(ElfBasicHeader)
        )

        return ProgramAttributes(
            elf_header.get_isa(),
            None,
            elf_basic_header.get_bitwidth(),
            elf_basic_header.get_endianness(),
            None,
        )


async def _create_deserializer(resource: Resource) -> BinaryDeserializer:
    elf_r = await resource.get_only_ancestor(ResourceFilter(tags=(Elf,)))
    e_basic_header = await elf_r.get_only_child_as_view(
        ElfBasicHeader, ResourceFilter.with_tags(ElfBasicHeader)
    )
    deserializer = BinaryDeserializer(
        io.BytesIO(await resource.get_data()),
        endianness=e_basic_header.get_endianness(),
        word_size=int(e_basic_header.get_bitwidth().get_word_size()),
    )
    return deserializer



import asyncio
import os.path
import tempfile
from dataclasses import dataclass
from subprocess import CalledProcessError

from ofrak.component.packer import Packer
from ofrak.component.unpacker import Unpacker, UnpackerError
from ofrak.resource import Resource
from ofrak.core.binary import GenericBinary
from ofrak.core.filesystem import FilesystemRoot, Folder, File, SpecialFileType
from ofrak.core.magic import MagicMimeIdentifier, MagicDescriptionIdentifier

from ofrak.model.component_model import CC, ComponentExternalTool
from ofrak_type.range import Range


TAR = ComponentExternalTool("tar", "https://www.gnu.org/software/tar/", "--help", apt_package="tar")


@dataclass
class TarArchive(GenericBinary, FilesystemRoot):
    """
    Filesystem stored in a tar archive.
    """


class TarUnpacker(Unpacker[None]):
    """
    Unpack a tar archive.
    """

    targets = (TarArchive,)
    children = (File, Folder, SpecialFileType)
    external_dependencies = (TAR,)

    async def unpack(self, resource: Resource, config: CC) -> None:
        # Write the archive data to a file
        with tempfile.NamedTemporaryFile(suffix=".tar") as temp_archive:
            temp_archive.write(await resource.get_data())
            temp_archive.flush()

            # Check the archive member files to ensure none unpack to a parent directory
            cmd = [
                "tar",
                "-P",
                "-tf",
                temp_archive.name,
            ]
            proc = await asyncio.create_subprocess_exec(
                *cmd,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            stdout, stderr = await proc.communicate()
            if proc.returncode:
                raise CalledProcessError(returncode=proc.returncode, cmd=cmd)
            for filename in stdout.decode().splitlines():
                # Handles relative parent paths and rooted paths, and normalizes paths like "./../"
                rel_filename = os.path.relpath(filename)
                if rel_filename.startswith("../"):
                    raise UnpackerError(
                        f"Tar archive contains a file {filename} that would extract to a parent "
                        f"directory {rel_filename}."
                    )

            # Unpack into a temporary directory using the temporary file
            with tempfile.TemporaryDirectory() as temp_dir:
                command = ["tar", "--xattrs", "-C", temp_dir, "-xf", temp_archive.name]
                proc = await asyncio.create_subprocess_exec(
                    *command,
                )
                returncode = await proc.wait()
                if returncode:
                    raise CalledProcessError(returncode=returncode, cmd=command)

                # Initialize a filesystem from the unpacked/untarred temporary folder
                tar_view = await resource.view_as(TarArchive)
                await tar_view.initialize_from_disk(temp_dir)


class TarPacker(Packer[None]):
    """
    Pack files into a tar archive.
    """

    targets = (TarArchive,)
    external_dependencies = (TAR,)

    async def pack(self, resource: Resource, config: CC) -> None:
        # Flush the child files to the filesystem
        tar_view = await resource.view_as(TarArchive)
        flush_dir = await tar_view.flush_to_disk()

        # Pack it back into a temporary archive
        with tempfile.NamedTemporaryFile(suffix=".tar") as temp_archive:
            cmd = [
                "tar",
                "--xattrs",
                "-C",
                flush_dir,
                "-cf",
                temp_archive.name,
                ".",
            ]
            proc = await asyncio.create_subprocess_exec(
                *cmd,
            )
            returncode = await proc.wait()
            if proc.returncode:
                raise CalledProcessError(returncode=returncode, cmd=cmd)

            # Replace the original archive data
            resource.queue_patch(Range(0, await resource.get_data_length()), temp_archive.read())


MagicMimeIdentifier.register(TarArchive, "application/x-tar")
MagicDescriptionIdentifier.register(TarArchive, lambda s: "tar archive" in s.lower())



import asyncio
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional, Iterable

from ofrak.resource import Resource

from ofrak.component.analyzer import Analyzer
from ofrak.component.unpacker import Unpacker
from ofrak.core.addressable import Addressable
from ofrak_type.architecture import InstructionSetMode
from ofrak.core.instruction import Instruction
from ofrak.core.memory_region import MemoryRegion
from ofrak.service.resource_service_i import ResourceFilter, ResourceSort, ResourceSortDirection


@dataclass
class BasicBlock(MemoryRegion):
    """
    A collection of [instructions][ofrak.core.instruction.Instruction] that constitute a single
    block of code with one entry point and one exit point.

    :ivar virtual_address: the virtual address corresponding ot the start of the basic block
    :ivar size: the basic block's size (the difference between the virtual address immediately after
    the block's last instruction and the block's virtual address)
    :ivar mode: the instruction set mode of the basic block
    :ivar is_exit_point: true if the basic block is the exit point of a function; if so,
    then exit_vaddr is None
    :ivar exit_vaddr: the virtual address of the next basic block (if the current basic block is
    not an exit point)
    """

    mode: InstructionSetMode
    is_exit_point: bool
    exit_vaddr: Optional[int]

    @classmethod
    def caption(cls, all_attributes) -> str:
        return str(cls.__name__)

    async def get_instructions(self) -> Iterable[Instruction]:
        """
        Get the basic block's [instructions][ofrak.core.instruction.Instruction].

        :return: the instructions in the basic block
        """
        return await self.resource.get_descendants_as_view(
            Instruction,
            r_filter=ResourceFilter.with_tags(Instruction),
            r_sort=ResourceSort(Addressable.VirtualAddress, ResourceSortDirection.ASCENDANT),
        )

    async def get_assembly(self) -> str:
        """
        Get the basic block's instructions as an assembly string.

        :return: the basic block's assembly
        """
        instructions = await self.get_instructions()
        instruction_assemblies = [i.get_assembly() for i in instructions]
        return "\n".join(await asyncio.gather(*instruction_assemblies))


class BasicBlockUnpacker(Unpacker[None], ABC):
    """
    Unpack a [basic block][ofrak.core.basic_block.BasicBlock] into
    [instructions][ofrak.core.instruction.Instruction].
    """

    id = b"BasicBlockUnpacker"
    targets = (BasicBlock,)
    children = (Instruction,)

    @abstractmethod
    async def unpack(self, resource: Resource, config=None):
        """
        Unpack a basic block into its corresponding instructions.
        """
        raise NotImplementedError()


class BasicBlockAnalyzer(Analyzer[None, BasicBlock], ABC):
    """
    Analyze a [basic block][ofrak.core.basic_block.BasicBlock] and extract its virtual address,
    size, mode, whether it is an exit point, and exit_vaddr.
    """

    targets = (BasicBlock,)
    outputs = (BasicBlock,)
    id = b"BasicBlockAnalyzer"

    @abstractmethod
    async def analyze(self, resource: Resource, config=None) -> BasicBlock:
        """
        Analyze a basic block and extract its virtual address, size, mode, whether it is an
        exit point, and exit_vaddr.

        :param resource: the basic block resource
        :param config:

        :return: the analyzed basic block
        """
        raise NotImplementedError()



import uuid

from ofrak.service.id_service_i import IDServiceInterface


class UUIDService(IDServiceInterface):
    def generate_id(self) -> bytes:
        # Note: pseudo-random
        return uuid.uuid4().bytes

    @staticmethod
    def generate_id_from_base(base_id: bytes, key: str) -> bytes:
        return uuid.uuid5(uuid.UUID(bytes=base_id), key).bytes



import io
import struct
import zlib
from dataclasses import dataclass
from enum import Enum
from typing import Optional, Tuple, Iterable, List

from ofrak.component.analyzer import Analyzer
from ofrak.component.modifier import Modifier
from ofrak.component.packer import Packer
from ofrak.component.unpacker import Unpacker
from ofrak.core import ProgramAttributes, GenericBinary, MagicDescriptionIdentifier
from ofrak.model.component_model import ComponentConfig
from ofrak.model.resource_model import ResourceAttributes
from ofrak.model.viewable_tag_model import AttributesType
from ofrak.resource import Resource
from ofrak.resource_view import ResourceView
from ofrak.service.resource_service_i import ResourceFilter
from ofrak_io.deserializer import BinaryDeserializer
from ofrak_type.architecture import InstructionSet
from ofrak_type.bit_width import BitWidth
from ofrak_type.endianness import Endianness
from ofrak_type.range import Range

UIMAGE_MAGIC = 0x27051956
UIMAGE_NAME_LEN = 32
UIMAGE_HEADER_LEN = 64  # Length of the UImage header in bytes


#####################
#       Enums       #
#####################
# Enums referenced from https://github.com/EmcraftSystems/u-boot/blob/master/include/image.h
class UImageType(Enum):
    INVALID = 0
    STANDALONE = 1
    KERNEL = 2
    RAMDISK = 3
    MULTI = 4
    FIRMWARE = 5
    SCRIPT = 6
    FILESYSTEM = 7
    FLATDT = 8
    KWBIMAGE = 9
    IMXIMAGE = 10


class UImageCompressionType(Enum):
    NONE = 0
    GZIP = 1
    BZIP2 = 2
    LZMA = 3
    LZO = 4


class UImageOperatingSystem(Enum):
    INVALID = 0
    OPENBSD = 1
    NETBSD = 2
    FREEBSD = 3
    BSD_4_4 = 4
    LINUX = 5
    SVR4 = 6
    ESIX = 7
    SOLARIS = 8
    IRIX = 9
    SCO = 10
    DELL = 11
    NCR = 12
    LYNXOS = 13
    VXWORKS = 14
    PSOS = 15
    QNX = 16
    U_BOOT = 17
    RTEMS = 18
    ARTOS = 19
    UNITY = 20
    INTEGRITY = 21


class UImageArch(Enum):
    INVALID = 0
    ALPHA = 1
    ARM = 2
    I386 = 3
    IA64 = 4
    MIPS = 5
    MIPS64 = 6
    PPC = 7
    S390 = 8
    SH = 9
    SPARC = 10
    SPARC64 = 11
    M68K = 12
    NIOS = 13
    MICROBLAZE = 14
    NIOS2 = 15
    BLACKFIN = 16
    AVR32 = 17
    ST200 = 18


#####################
#     RESOURCES     #
#####################
@dataclass
class UImageHeader(ResourceView):
    """UImage header.

    :var ih_magic: Image Header Magic Number, 4 bytes
    :var ih_hcrc: Image Header CRC Checksum (when this field itself is zeroed out), 4 bytes
    :var ih_time: Image Creation Timestamp, 4 bytes
    :var ih_size: Image Data Size, 4 bytes
    :var ih_load: Data Load  Address, 4 bytes
    :var ih_ep: Entry Point Address, 4 bytes
    :var ih_dcrc: Image Data CRC Checksum, 4 bytes
    :var ih_os: Operating System, 1 byte
    :var ih_arch: CPU architecture, 1 byte
    :var ih_type: Image Type, 1 byte
    :var ih_comp: Compression Type, 1 byte
    :var ih_name: Image Name, 32 bytes
    """

    ih_magic: int
    ih_hcrc: int
    ih_time: int
    ih_size: int
    ih_load: int
    ih_ep: int
    ih_dcrc: int
    ih_os: int
    ih_arch: int
    ih_type: int
    ih_comp: int
    ih_name: bytes

    def get_os(self) -> UImageOperatingSystem:
        return UImageOperatingSystem(self.ih_os)

    def get_arch(self) -> UImageArch:
        return UImageArch(self.ih_arch)

    def get_compression_type(self) -> UImageCompressionType:
        return UImageCompressionType(self.ih_comp)

    def get_type(self) -> UImageType:
        return UImageType(self.ih_type)

    def get_name(self) -> str:
        return str(self.ih_name)

    def get_data_size(self) -> int:
        return self.ih_size

    def get_load_vaddr(self) -> int:
        return self.ih_load

    def get_entry_point_vaddr(self) -> int:
        return self.ih_ep


@dataclass
class UImageMultiHeader(ResourceView):
    """
    UImage MULTI-type header holding sizes of bodies contained in the parent uimage
    The header size is unknown, and is calculated by the uboot loader by reading past the
    UImageHeader until it reaches a null dword (\x00\x00\x00\x00).
    """

    image_sizes: Tuple[int, ...]

    def get_image_sizes(self) -> Tuple[int, ...]:
        return self.image_sizes

    def get_number_of_bodies(self) -> int:
        return len(self.image_sizes)

    def get_multi_header_size(self) -> int:
        return (len(self.image_sizes) + 1) * 4


class UImageBody(GenericBinary):
    pass


class UImageTrailingBytes(GenericBinary):
    pass


@dataclass
class UImage(GenericBinary):
    """
    A UImage is an image file that has a U-Boot wrapper (installed by the mkimage utility), and
    contains 1 or more images for use by U-Boot during boot.

    The U-Boot wrapper, or UImageHeader is a 64-byte fixed size header which contains
    information about the included images, such as the OS type, loader information, hardware ISA,
    etc. as well as CRC32 integrity checks.

    The contained images, or UImageBodies, can be anything from other nested UImages, to
    filesystems, kernels, and scripts.

    Assuming the body of a UImage is a program, the following snippet outlines a way to tag and
    analyze it.
    Note the addition of a CodeRegion view with a virtual address obtained from the UImageHeader,
    as well as tagging the UImageBody as a Program.
    ```
    uimage = await resource.view_as(UImage)
    uimage_header = await uimage.get_header()
    uimage_bodies = await uimage.get_bodies()
    uimage_body = uimage_bodies[0]
    uimage_body.resource.add_view(
        CodeRegion(
            virtual_address=uimage_header.get_load_vaddr(),
            size=uimage_header.get_data_size(),
        )
    )
    uimage_body.resource.add_tag(Program)
    await uimage_body.resource.save()
    ```
    """

    async def get_header(self) -> UImageHeader:
        return await self.resource.get_only_child_as_view(
            UImageHeader,
            ResourceFilter.with_tags(UImageHeader),
        )

    async def get_multi_header(self) -> UImageMultiHeader:
        return await self.resource.get_only_child_as_view(
            UImageMultiHeader, ResourceFilter.with_tags(UImageMultiHeader)
        )

    async def get_bodies(self) -> Iterable[UImageBody]:
        return await self.resource.get_children_as_view(
            UImageBody,
            ResourceFilter.with_tags(UImageBody),
        )


####################
#    UNPACKERS     #
####################
class UImageUnpacker(Unpacker[None]):
    """
    UImage unpacker.

    It unpacks the UImage resource into:
     - A 64-byte UImageHeader
     - [Optional] UImageMultiHeader (for UImageTypes.MULTI images)
     - A list of UImageBody instances (1 or many)
     - [Optional] UImageTrailingBytes (if any)
    """

    id = b"UImageUnpacker"
    targets = (UImage,)
    children = (
        UImageHeader,
        UImageMultiHeader,
        UImageBody,
        UImageTrailingBytes,
    )

    async def unpack(self, resource: Resource, config=None):
        uimage_header_r = await resource.create_child(
            tags=(UImageHeader,), data_range=Range(0, UIMAGE_HEADER_LEN)
        )
        uimage_header = await uimage_header_r.view_as(UImageHeader)
        resource_data = await resource.get_data()
        if uimage_header.get_type() == UImageType.MULTI:
            uimage_multi_size = resource_data[UIMAGE_HEADER_LEN:].find(b"\x00" * 4) + 4
            await resource.create_child(
                tags=(UImageMultiHeader,),
                data=resource_data[UIMAGE_HEADER_LEN : UIMAGE_HEADER_LEN + uimage_multi_size],
            )
            uimage_multi_header = await resource.get_only_child_as_view(
                UImageMultiHeader, ResourceFilter.with_tags(UImageMultiHeader)
            )

            image_i_start = UIMAGE_HEADER_LEN + uimage_multi_size
            for image_size in uimage_multi_header.get_image_sizes():
                await resource.create_child(
                    tags=(UImageBody,),
                    data=resource_data[image_i_start : image_i_start + image_size],
                )
                image_i_start += image_size

            total_len_with_bodies = (
                UIMAGE_HEADER_LEN
                + uimage_multi_header.get_multi_header_size()
                + sum(uimage_multi_header.get_image_sizes())
            )
            if total_len_with_bodies < uimage_header.ih_size:
                await resource.create_child(
                    tags=(UImageTrailingBytes,), data=resource_data[total_len_with_bodies:]
                )
        else:
            await resource.create_child(
                tags=(UImageBody,),
                data=resource_data[UIMAGE_HEADER_LEN:],
            )


#####################
#     ANALYZERS     #
#####################
class UImageHeaderAttributesAnalyzer(Analyzer[None, UImageHeader]):
    """
    Analyze the UImageHeader of a UImage
    """

    targets = (UImageHeader,)
    outputs = (UImageHeader,)

    async def analyze(self, resource: Resource, config=None) -> UImageHeader:
        tmp = await resource.get_data()
        deserializer = BinaryDeserializer(
            io.BytesIO(tmp),
            endianness=Endianness.BIG_ENDIAN,
            word_size=4,
        )

        deserialized = deserializer.unpack_multiple(f"IIIIIIIBBBB{UIMAGE_NAME_LEN}s")
        (
            ih_magic,
            ih_hcrc,
            ih_time,
            ih_size,
            ih_load,
            ih_ep,
            ih_dcrc,
            ih_os,
            ih_arch,
            ih_type,
            ih_comp,
            ih_name,
        ) = deserialized

        assert ih_magic == UIMAGE_MAGIC

        return UImageHeader(
            ih_magic,
            ih_hcrc,
            ih_time,
            ih_size,
            ih_load,
            ih_ep,
            ih_dcrc,
            ih_os,
            ih_arch,
            ih_type,
            ih_comp,
            ih_name,
        )


class UImageMultiHeaderAttributesAnalyzer(Analyzer[None, UImageMultiHeader]):
    """
    Analyze the UImageMultiHeader of a UImageType.MULTI UImage
    """

    targets = (UImageMultiHeader,)
    outputs = (UImageMultiHeader,)

    async def analyze(self, resource: Resource, config=None) -> UImageMultiHeader:
        resource_data = await resource.get_data()
        deserializer = BinaryDeserializer(
            io.BytesIO(resource_data),
            endianness=Endianness.BIG_ENDIAN,
            word_size=4,
        )
        uimage_multi_header_size = (len(resource_data) - 4) // 4  # Remove trailing null dword
        deserialized = deserializer.unpack_multiple(f"{uimage_multi_header_size}I")
        return UImageMultiHeader(deserialized)


class UImageProgramAttributesAnalyzer(Analyzer[None, Tuple[ProgramAttributes]]):
    """
    Analyze the ProgramAttributes of a UImage from its header
    """

    targets = (UImage,)
    outputs = (ProgramAttributes,)

    async def analyze(self, resource: Resource, config=None) -> Tuple[ProgramAttributes]:
        uimage_view = await resource.view_as(UImage)
        uimage_header = await uimage_view.get_header()
        uimage_program_attributes = self.from_deserialized_header(uimage_header)
        return (uimage_program_attributes,)

    @staticmethod
    def from_deserialized_header(
        header: UImageHeader,
    ) -> ProgramAttributes:
        UIMAGE_ARCH_TO_ISA = {
            UImageArch.ARM: InstructionSet.ARM,
            UImageArch.I386: InstructionSet.X86,
            UImageArch.IA64: InstructionSet.X86,
            UImageArch.MIPS: InstructionSet.MIPS,
            UImageArch.MIPS64: InstructionSet.MIPS,
            UImageArch.PPC: InstructionSet.PPC,
        }

        UIMAGE_ARCH_TO_BIT_WIDTH = {
            UImageArch.ARM: BitWidth.BIT_32,
            UImageArch.I386: BitWidth.BIT_32,
            UImageArch.IA64: BitWidth.BIT_64,
            UImageArch.MIPS: BitWidth.BIT_32,
            UImageArch.MIPS64: BitWidth.BIT_64,
            UImageArch.PPC: BitWidth.BIT_32,
        }

        UIMAGE_ARCH_TO_ENDIANNESS = {
            UImageArch.ARM: Endianness.LITTLE_ENDIAN,
            UImageArch.I386: Endianness.LITTLE_ENDIAN,
            UImageArch.IA64: Endianness.LITTLE_ENDIAN,
            UImageArch.PPC: Endianness.BIG_ENDIAN,
        }

        uimage_arch = UImageArch(header.ih_arch)

        try:
            isa = UIMAGE_ARCH_TO_ISA[uimage_arch]
            bit_width = UIMAGE_ARCH_TO_BIT_WIDTH[uimage_arch]
            endianness = UIMAGE_ARCH_TO_ENDIANNESS[uimage_arch]
        except ValueError:
            raise NotImplementedError(
                f"Unsupported/unknown uImage architecture: {uimage_arch.name}"
            )

        return ProgramAttributes(isa, None, bit_width, endianness, None)


####################
#    MODIFIERS     #
####################
@dataclass
class UImageHeaderModifierConfig(ComponentConfig):
    """
    Modifier config for a UImageHeader.

    The following field is not modifiable:
    - Image Header Magic Number (a constant)

    The following field is modifiable, but will be overwritten by the modifier:
    - Image Header CRC Checksum (`ih_hcrc`)

    The following fields are modifiable, but will be overwritten by the packer:
    - Image Data Size (`ih_size`)
    - Image Data CRC Checksum (`ih_dcrc`)
    """

    ih_hcrc: Optional[int] = None
    ih_time: Optional[int] = None
    ih_size: Optional[int] = None
    ih_load: Optional[int] = None
    ih_ep: Optional[int] = None
    ih_dcrc: Optional[int] = None
    ih_os: Optional[int] = None
    ih_arch: Optional[int] = None
    ih_type: Optional[int] = None
    ih_comp: Optional[int] = None
    ih_name: Optional[bytes] = None


class UImageHeaderModifier(Modifier[UImageHeaderModifierConfig]):
    """
    Modify a UImageHeader according to a given modifier config.
    Updates the header CRC field (`ih_hcrc`).
    """

    targets = (UImageHeader,)

    async def modify(self, resource: Resource, config: UImageHeaderModifierConfig) -> None:
        original_attributes = await resource.analyze(AttributesType[UImageHeader])
        # First serialize the header with the ih_hcrc field set to 0, to compute this CRC later
        new_attributes = ResourceAttributes.replace_updated(original_attributes, config)
        tmp_serialized_header = await UImageHeaderModifier.serialize(new_attributes, ih_hcrc=0)
        # Patch this header with its CRC32 in the ih_hcrc field
        ih_hcrc = zlib.crc32(tmp_serialized_header)
        serialized_header = await UImageHeaderModifier.serialize(new_attributes, ih_hcrc=ih_hcrc)
        resource.queue_patch(Range.from_size(0, UIMAGE_HEADER_LEN), serialized_header)
        new_attributes = ResourceAttributes.replace_updated(
            new_attributes, UImageHeaderModifierConfig(ih_hcrc=ih_hcrc)
        )
        resource.add_attributes(new_attributes)

    @staticmethod
    async def serialize(
        updated_attributes: AttributesType[UImageHeader], ih_hcrc: int = 0
    ) -> bytes:
        """
        Serialize `updated_attributes` into bytes, using `ih_hcrc` for the eponymous field.
        This method doesn't perform any check or compute any CRC.
        """
        return struct.pack(
            f"!IIIIIIIBBBB{UIMAGE_NAME_LEN}s",
            UIMAGE_MAGIC,
            ih_hcrc,
            updated_attributes.ih_time,
            updated_attributes.ih_size,
            updated_attributes.ih_load,
            updated_attributes.ih_ep,
            updated_attributes.ih_dcrc,
            updated_attributes.ih_os,
            updated_attributes.ih_arch,
            updated_attributes.ih_type,
            updated_attributes.ih_comp,
            updated_attributes.ih_name,
        )


@dataclass
class UImageMultiHeaderModifierConfig(ComponentConfig):
    """
    Modifier config for a UImageMultiHeader.
    """

    image_sizes: List[int]


class UImageMultiHeaderModifier(Modifier[UImageMultiHeaderModifierConfig]):
    """
    Modify a UImageMultiHeader according to a given modifier config.
    """

    targets = (UImageMultiHeader,)

    async def modify(self, resource: Resource, config: UImageMultiHeaderModifierConfig) -> None:

        # # First serialize the header with the ih_hcrc field set to 0, to compute this CRC later
        original_attributes = await resource.analyze(AttributesType[UImageMultiHeader])
        new_attributes = ResourceAttributes.replace_updated(original_attributes, config)
        serialized_multiheader = await UImageMultiHeaderModifier.serialize(new_attributes)
        uimage_multi_header = await resource.view_as(UImageMultiHeader)
        resource.queue_patch(
            Range(0, uimage_multi_header.get_multi_header_size()),
            serialized_multiheader,
        )
        resource.add_attributes(new_attributes)

    @staticmethod
    async def serialize(
        updated_attributes: AttributesType[UImageMultiHeader],
    ) -> bytes:
        """
        Serialize `updated_attributes` into bytes
        """
        serialized = b""
        for image_size in updated_attributes.image_sizes:
            serialized += struct.pack("!I", image_size)
        serialized += b"\x00" * 4
        return serialized


####################
#     PACKERS      #
####################
class UImagePacker(Packer[None]):
    """
    UImage packer.

    It patches the resource's UImageHeader and UImageBody instances into a single binary,
    updating the CRC checksums and image data size in the UImageHeader.
    Also handles the UImageMultiHeader in the case of UImageTypes.MULTI UImages.
    """

    id = b"UImagePacker"
    targets = (UImage,)

    async def pack(self, resource: Resource, config=None):
        repacked_body_data = b""
        uimage_view = await resource.view_as(UImage)
        header = await uimage_view.get_header()
        if header.get_type() == UImageType.MULTI:
            image_sizes = []
            for uimage_body in await uimage_view.get_bodies():
                image_sizes.append(await uimage_body.resource.get_data_length())
            multi_header = await uimage_view.get_multi_header()
            multiheader_modifier_config = UImageMultiHeaderModifierConfig(image_sizes=image_sizes)
            await multi_header.resource.run(UImageMultiHeaderModifier, multiheader_modifier_config)
            repacked_body_data += await multi_header.resource.get_data()
        for uimage_body in await uimage_view.get_bodies():
            repacked_body_data += await uimage_body.resource.get_data()

        # If there are UImageTrailingBytes, get them as well.
        resource_children = await resource.get_children()
        if any([child.has_tag(UImageTrailingBytes) for child in resource_children]):
            trailing_bytes_r = await resource.get_only_child_as_view(
                UImageTrailingBytes, ResourceFilter.with_tags(UImageTrailingBytes)
            )
            repacked_body_data += await trailing_bytes_r.resource.get_data()
        ih_size = len(repacked_body_data)
        ih_dcrc = zlib.crc32(repacked_body_data)
        header_modifier_config = UImageHeaderModifierConfig(ih_size=ih_size, ih_dcrc=ih_dcrc)
        await header.resource.run(UImageHeaderModifier, header_modifier_config)
        # Patch UImageHeader data
        header_data = await header.resource.get_data()
        # Patch all other data
        original_size = await resource.get_data_length()
        resource.queue_patch(Range(0, original_size), header_data + repacked_body_data)


MagicDescriptionIdentifier.register(UImage, lambda s: s.startswith("u-boot legacy uImage"))



"""
This component is intended to make it easier to analyze raw flash dumps using OFRAK alone.
Most flash dumps have "useful" data mixed in with out-of-band (OOB) data.
The OOB data often includes some Error Correcting Codes (ECC) or checksums.

There are several dataclasses that categorize the sections of the dump:

- `FlashResource` is the overarching resource. The component expects the user to add this tag in order for this component to be run.
    - `FlashOobResource` is the region of `FlashResource` that has OOB data. In the future, there may be multiple of these children resources.
        - `FlashLogicalDataResource` is the extracted data only with all of the OOB data removed. This will become a `FlashOobResource` when packed.
        - `FlashLogicalEccResource` is the extracted ECC only. No other OOB data is included.
"""
import logging
from dataclasses import dataclass
from enum import Enum
from hashlib import md5
from typing import Any, Callable, Dict, Generator, Iterable, Optional

from ofrak.component.packer import Packer, PackerError
from ofrak.component.unpacker import Unpacker, UnpackerError
from ofrak.core.binary import GenericBinary
from ofrak.core.ecc.abstract import EccAlgorithm, EccError
from ofrak.model.resource_model import ResourceAttributes
from ofrak.resource import Resource
from ofrak.service.resource_service_i import ResourceFilter
from ofrak_type.error import NotFoundError
from ofrak_type.range import Range

LOGGER = logging.getLogger()

# Dict of data mapping MD5 checksum to ECC bytes, used to check for updates
DATA_HASHES: Dict[bytes, bytes] = dict()


#####################
#     RESOURCES     #
#####################
@dataclass
class FlashResource(GenericBinary):
    """
    The overarching resource that encapsulates flash storage.
    This will contain a `FlashOobResource` in most cases.
    In the future, support for multiple `FlashOobResource` children should be added.
    """


@dataclass
class FlashOobResource(GenericBinary):
    """
    Represents the region containing Oob data.
    """


@dataclass
class FlashLogicalDataResource(GenericBinary):
    """
    This is the final product of unpacking a FlashResource.
    It contains the data without any ECC or OOB data included.
    This allows for recursive packing and unpacking.
    """


@dataclass
class FlashLogicalEccResource(GenericBinary):
    """
    The alternate to FlashLogicalDataResource but just includes ECC.
    Does not include any other OOB data.
    Generally less useful on its own but provided anyway.
    """


#####################
#     ATTRIBUTES    #
#####################
@dataclass(**ResourceAttributes.DATACLASS_PARAMS)
class FlashEccAttributes(ResourceAttributes):
    """
    Must be configured if the resource includes ECC
    `ecc_magic` is assumed to be contained at the start of the file, but may also occur multiple times
    """

    ecc_class: Optional[EccAlgorithm] = None
    ecc_magic: Optional[bytes] = None
    head_delimiter: Optional[bytes] = None
    first_data_delimiter: Optional[bytes] = None
    data_delimiter: Optional[bytes] = None
    last_data_delimiter: Optional[bytes] = None
    tail_delimiter: Optional[bytes] = None


class FlashFieldType(Enum):
    """
    `DATA_SIZE` is the packed size of the DATA only (excluding `MAGIC`, `CHECKSUM`, `DELIMITER`, `ECC`, etc)
    `TOTAL_SIZE` is the size of the entire region (including all `DATA`, `MAGIC`, `CHECKSUM`, `DELIMITER`, `ECC`, etc)
    `ALIGNMENT` will pad with \x00 bytes by default
    """

    DATA = 0
    ECC = 1
    ALIGNMENT = 2
    MAGIC = 3
    DATA_SIZE = 4
    ECC_SIZE = 5
    CHECKSUM = 6
    DELIMITER = 7
    TOTAL_SIZE = 8


@dataclass
class FlashField:
    field_type: FlashFieldType
    size: int


@dataclass(**ResourceAttributes.DATACLASS_PARAMS)
class FlashAttributes(ResourceAttributes):
    """
    FlashAttributes is for specifying everything about the specific model of flash.
    The intent is to expand to all common flash configurations.
    Every block has a format specifier to show where each field is in the block as well as the length.
    If there is no OOB data, data_block_format may take this form:

        data_block_format = [FlashField(field_type=FlashFieldType.DATA,size=block_size),]

    Important Notes:
    Assumes that the provided list for each block format is ordered.
    Only define a block_format if they are different from other block formats.
        - A current workaround is adding `FlashField(FlashFieldType.ALIGNMENT, 0)`
    Assumes that there are only one of each block format except the data_block_format
    """

    data_block_format: Iterable[FlashField]
    header_block_format: Optional[Iterable[FlashField]] = None
    first_data_block_format: Optional[Iterable[FlashField]] = None
    last_data_block_format: Optional[Iterable[FlashField]] = None
    tail_block_format: Optional[Iterable[FlashField]] = None
    ecc_attributes: Optional[FlashEccAttributes] = None
    checksum_func: Optional[Callable[[Any], Any]] = None

    def get_block_formats(self) -> Iterable:
        return filter(
            None,
            [
                self.header_block_format,
                self.first_data_block_format,
                self.data_block_format,
                self.last_data_block_format,
                self.tail_block_format,
            ],
        )

    def get_block_size(self, block_format: Iterable[FlashField]) -> int:
        if block_format is not None:
            return sum(field.size for field in block_format)
        else:
            return 0

    def get_oob_size_in_block(self, block_format: Iterable[FlashField]) -> int:
        if block_format is None:
            return 0
        data_length = self.get_field_length_in_block(
            block_format=block_format, field_type=FlashFieldType.DATA
        )
        return self.get_block_size(block_format=block_format) - data_length

    def get_field_in_block(
        self, block_format: Optional[Iterable[FlashField]], field_type: FlashFieldType
    ) -> Optional[FlashField]:
        if block_format is None:
            return None
        for field in block_format:
            if field.field_type is field_type:
                return field
        return None

    def get_field_range_in_block(
        self, block_format: Optional[Iterable[FlashField]], field_type: FlashFieldType
    ) -> Optional[Range]:
        offset = 0
        if block_format is None:
            return None
        for field in block_format:
            if field.field_type is field_type:
                data_size = field.size
                return Range(offset, offset + data_size)
            # Add all data in fields that come before data
            offset += field.size
        return None

    def get_field_length_in_block(
        self, block_format: Iterable[FlashField], field_type: FlashFieldType
    ) -> int:
        if block_format is None:
            return 0
        field_range = self.get_field_range_in_block(
            block_format=block_format, field_type=field_type
        )
        if field_range is not None:
            return field_range.length()
        return 0

    def get_field_data_in_block(
        self,
        block_format: Optional[Iterable[FlashField]],
        field_type: FlashFieldType,
        data: bytes,
        block_start_offset: int,
    ) -> Optional[bytes]:
        field_range = self.get_field_range_in_block(
            block_format=block_format, field_type=field_type
        )
        if field_range is not None:
            return data[
                block_start_offset + field_range.start : block_start_offset + field_range.end
            ]
        return None

    def get_num_data_blocks(self, data_len: int, includes_oob: bool = False) -> int:
        data_block_count = 0
        data_count = 0
        for c in self.get_block_formats():
            if c != self.data_block_format:
                # Skip data block for now
                if includes_oob:
                    block_data_len = self.get_block_size(c)
                else:
                    block_data_len = self.get_field_length_in_block(c, FlashFieldType.DATA)

                if block_data_len is not None:
                    data_count += block_data_len

        if includes_oob:
            block_data_len = self.get_block_size(self.data_block_format)
        else:
            block_data_len = self.get_field_length_in_block(
                self.data_block_format, FlashFieldType.DATA
            )

        while data_count < data_len:
            # The rest of the blocks are data blocks
            data_count += block_data_len
            data_block_count += 1
        return data_block_count

    def iterate_through_all_blocks(
        self, data_len: int, includes_oob: bool
    ) -> Generator[Iterable[FlashField], None, int]:
        count = 0
        for c in self.get_block_formats():
            num_blocks = (
                self.get_num_data_blocks(data_len, includes_oob)
                if c is self.data_block_format
                else 1
            )
            for _ in range(0, num_blocks):
                yield c
                count += 1
        return count

    def get_total_oob_size(self, data_len: int, includes_oob: bool = False) -> int:
        total_oob_size = 0
        for c in self.get_block_formats():
            block_oob_size = self.get_oob_size_in_block(c)
            if block_oob_size is not None:
                num_blocks = (
                    self.get_num_data_blocks(data_len, includes_oob)
                    if c is self.data_block_format
                    else 1
                )
                total_oob_size += block_oob_size * num_blocks
        return total_oob_size

    def get_total_field_size(
        self, data_len: int, field_type: FlashFieldType, includes_oob: bool = False
    ) -> int:
        total_field_size = 0
        for c in self.get_block_formats():
            block_field_size = self.get_field_length_in_block(c, field_type)
            if block_field_size is not None:
                num_blocks = (
                    self.get_num_data_blocks(data_len, includes_oob)
                    if c is self.data_block_format
                    else 1
                )
                total_field_size += block_field_size * num_blocks
        return total_field_size


#####################
#     UNPACKERS     #
#####################
class FlashResourceUnpacker(Unpacker[None]):
    """
    Finds the overarching parent for region that includes OOB data.
    Identifies the bounds based on the `FlashAttributes`.
    """

    targets = (FlashResource,)
    children = (FlashOobResource,)

    async def unpack(self, resource: Resource, config=None):
        try:
            flash_attr = resource.get_attributes(FlashAttributes)
        except NotFoundError:
            raise UnpackerError("Tried creating FlashOobResource without FlashAttributes")

        start_index = 0
        data = await resource.get_data()
        data_len = len(data)

        if flash_attr.ecc_attributes is not None:
            magic = flash_attr.ecc_attributes.ecc_magic
            if magic is not None:
                ecc_magic_offset = data.find(magic)

                if ecc_magic_offset == -1:
                    raise UnpackerError("No header magic found")

                magic_range_in_block = flash_attr.get_field_range_in_block(
                    flash_attr.header_block_format, FlashFieldType.MAGIC
                )

                if not magic_range_in_block:
                    raise UnpackerError("Did not find offset for magic in header")

                start_index = ecc_magic_offset - magic_range_in_block.start

        # Set fallback, in case the current check for the end of the resource fails
        end_offset = data_len
        header_data_size = flash_attr.get_field_in_block(
            flash_attr.header_block_format, FlashFieldType.DATA_SIZE
        )
        header_total_size = flash_attr.get_field_in_block(
            flash_attr.header_block_format, FlashFieldType.TOTAL_SIZE
        )
        if flash_attr.header_block_format is not None and (
            header_data_size is not None or header_total_size is not None
        ):
            # The header has the size of the entire region including OOB data
            total_ecc_protected_size = 0
            if header_data_size is not None:
                # Found data size in the header, need to calculate expected total size (including OOB)
                data_size_bytes = flash_attr.get_field_data_in_block(
                    block_format=flash_attr.header_block_format,
                    field_type=FlashFieldType.DATA_SIZE,
                    data=data,
                    block_start_offset=0,
                )
                oob_size = flash_attr.get_total_oob_size(data_len=data_len, includes_oob=True)
                if data_size_bytes is not None:
                    total_ecc_protected_size = oob_size + int.from_bytes(data_size_bytes, "big")
            elif header_total_size is not None:
                # Found total size in header
                total_size_bytes = flash_attr.get_field_data_in_block(
                    block_format=flash_attr.header_block_format,
                    field_type=FlashFieldType.TOTAL_SIZE,
                    data=data,
                    block_start_offset=0,
                )
                if total_size_bytes is not None:
                    total_ecc_protected_size = int.from_bytes(total_size_bytes, "big")

            if total_ecc_protected_size > data_len:
                raise UnpackerError("Expected larger resource than supplied")

            if total_ecc_protected_size > start_index:
                end_offset = start_index + total_ecc_protected_size

        elif flash_attr.tail_block_format is not None:
            # Tail has either magic, total_size, or data_size to indicate the end
            tail_magic = flash_attr.get_field_in_block(
                flash_attr.tail_block_format, FlashFieldType.MAGIC
            )
            tail_total_size = flash_attr.get_field_in_block(
                flash_attr.tail_block_format, FlashFieldType.TOTAL_SIZE
            )
            tail_data_size = flash_attr.get_field_in_block(
                flash_attr.tail_block_format, FlashFieldType.DATA_SIZE
            )

            if tail_magic is not None:
                # we'll start looking after the header to make sure we don't
                # accidentally find the header magic
                if flash_attr.header_block_format:
                    tail_start_index = flash_attr.get_block_size(flash_attr.header_block_format)
                else:
                    tail_start_index = start_index
                end_offset = _get_end_from_magic(flash_attr, tail_start_index, data, data_len)
            elif tail_total_size is not None:
                end_offset = _get_end_from_size(
                    flash_attr, start_index, data, data_len, FlashFieldType.TOTAL_SIZE
                )
            elif tail_data_size is not None:
                end_offset = _get_end_from_size(
                    flash_attr, start_index, data, data_len, FlashFieldType.DATA_SIZE
                )

        # Create the overarching resource
        return await resource.create_child(
            tags=(FlashOobResource,),
            data_range=Range(start_index, end_offset),
            attributes=[
                flash_attr,
            ],
        )


class FlashOobResourceUnpacker(Unpacker[None]):
    """
    Unpack a single `FlashOobResource` dump into logical data using the `FlashAttributes`.
    """

    targets = (FlashOobResource,)
    children = (
        FlashLogicalDataResource,
        FlashLogicalEccResource,
    )

    async def unpack(self, resource: Resource, config=None):
        try:
            flash_attr = resource.get_attributes(FlashAttributes)
        except NotFoundError:
            raise UnpackerError("Tried unpacking without FlashAttributes")
        ecc_attr: Optional[FlashEccAttributes] = flash_attr.ecc_attributes

        oob_resource = resource
        # Parent FlashEccResource is created, redefine data to limited scope
        data = await oob_resource.get_data()
        data_len = len(data)

        # Now add children blocks until we reach the tail block
        offset = 0
        only_data = list()
        only_ecc = list()
        for block in flash_attr.iterate_through_all_blocks(data_len, True):
            block_size = flash_attr.get_block_size(block)
            block_end_offset = offset + block_size
            if block_end_offset > data_len:
                LOGGER.info(
                    f"Block offset {block_end_offset} is {block_end_offset - data_len} larger "
                    f"than {data_len}. In this case unpacking is best effort and end of unpacked "
                    f"child might not be accurate."
                )
                break
            block_range = Range(offset, block_end_offset)
            block_data = await oob_resource.get_data(range=block_range)

            # Iterate through every field in block, dealing with ECC and DATA
            block_ecc_range = None
            block_data_range = None
            field_offset = 0
            for field_index, field in enumerate(block):
                field_range = Range(field_offset, field_offset + field.size)

                # We must check all blocks anyway so deal with ECC here
                if field.field_type == FlashFieldType.ECC:
                    block_ecc_range = field_range
                    cur_block_ecc = block_data[block_ecc_range.start : block_ecc_range.end]
                    only_ecc.append(cur_block_ecc)
                    # Add hash of everything up to the ECC to our dict for faster packing
                    block_data_hash = md5(block_data[: block_ecc_range.start]).digest()
                    DATA_HASHES[block_data_hash] = cur_block_ecc

                if field.field_type == FlashFieldType.DATA:
                    block_data_range = field_range
                    # Get next ECC range
                    future_offset = field_offset
                    block_list = list(block)
                    for future_field in block_list[field_index:]:
                        if future_field.field_type == FlashFieldType.ECC:
                            block_ecc_range = Range(
                                future_offset, future_offset + future_field.size
                            )
                        future_offset += future_field.size

                    if block_ecc_range is not None:
                        # Try decoding/correcting with ECC, report any error
                        try:
                            # Assumes that data comes before ECC
                            if (ecc_attr is not None) and (ecc_attr.ecc_class is not None):
                                only_data.append(
                                    ecc_attr.ecc_class.decode(block_data[: block_ecc_range.end])[
                                        block_data_range.start : block_data_range.end
                                    ]
                                )
                            else:
                                raise UnpackerError(
                                    "Tried to correct with ECC without providing an ecc_class in FlashEccAttributes"
                                )
                        except EccError:
                            raise UnpackerError("ECC correction failed")
                    else:
                        # No ECC found in block, just add the data directly
                        only_data.append(block_data[block_data_range.start : block_data_range.end])
                field_offset += field.size
            offset += block_size

        # Add all block data to logical resource for recursive unpacking
        await oob_resource.create_child(
            tags=(FlashLogicalDataResource,),
            data=b"".join(only_data) if only_data else data,
            attributes=[
                flash_attr,
            ],
        )
        if ecc_attr is not None:
            await oob_resource.create_child(
                tags=(FlashLogicalEccResource,),
                data=b"".join(only_ecc),
                attributes=[
                    ecc_attr,
                ],
            )


#####################
#      PACKERS      #
#####################
class FlashResourcePacker(Packer[None]):
    """
    Packs the FlashResource into binary and cleans up logical data representations
    """

    id = b"FlashResourcePacker"
    targets = (FlashResource,)

    async def pack(self, resource: Resource, config=None):
        # We want to overwrite ourselves with just the repacked version
        # TODO: Add supoort for multiple FlashOobResource in a dump.
        packed_child = await resource.get_only_child(
            r_filter=ResourceFilter.with_tags(
                FlashOobResource,
            ),
        )
        if packed_child is not None:
            patch_data = await packed_child.get_data()
            original_size = await resource.get_data_length()
            resource.queue_patch(Range(0, original_size), patch_data)


class FlashOobResourcePacker(Packer[None]):
    """
    Packs the entire region including Oob data back into a binary blob
    """

    id = b"FlashOobResourcePacker"
    targets = (FlashOobResource,)

    async def pack(self, resource: Resource, config=None):
        # We want to overwrite ourselves with just the repacked version
        packed_child = await resource.get_only_child(
            r_filter=ResourceFilter.with_tags(
                FlashOobResource,
            ),
        )
        if packed_child is not None:
            patch_data = await packed_child.get_data()
            original_size = await resource.get_data_length()
            resource.queue_patch(Range(0, original_size), patch_data)


class FlashLogicalDataResourcePacker(Packer[None]):
    """
    Packs the `FlashLogicalDataResource` into a `FlashOobResource` of the format
    specified by the `FlashAttributes`
    """

    id = b"FlashLogicalDataResourcePacker"
    targets = (FlashLogicalDataResource,)

    async def pack(self, resource: Resource, config=None):
        try:
            flash_attr = resource.get_attributes(FlashAttributes)
        except NotFoundError:
            raise UnpackerError("Tried packing without FlashAttributes")
        data = await resource.get_data()
        bytes_left = len(data)
        original_size = bytes_left
        packed_data = bytearray()
        data_offset = 0

        for c in flash_attr.iterate_through_all_blocks(original_size, False):
            block_data = b""
            block_data_size = flash_attr.get_field_length_in_block(c, FlashFieldType.DATA)
            if block_data_size != 0:
                # Get the data for the current block
                block_data = data[data_offset : data_offset + block_data_size]
                data_offset += block_data_size
                bytes_left -= block_data_size

            packed_data += _build_block(
                cur_block_type=c,
                attributes=flash_attr,
                block_data=block_data,
                original_data=data,
            )

        # Create child under the original FlashOobResource to show that it packed itself
        parent = await resource.get_parent()
        await parent.create_child(tags=(FlashOobResource,), data=packed_data)


#####################
#      HELPERS      #
#####################
def _get_end_from_magic(attributes: FlashAttributes, start_index: int, data: bytes, data_len: int):
    search_offset = start_index
    search_field = FlashFieldType.MAGIC
    if (
        attributes is None
        or attributes.tail_block_format is None
        or attributes.ecc_attributes is None
    ):
        raise UnpackerError("Tried to find magic without all attributes defined")

    search_key = attributes.ecc_attributes.ecc_magic
    if search_key is None:
        raise UnpackerError(
            "Tried to find magic in tail without providing attribute in FlashEccAttributes"
        )

    while 0 <= search_offset <= data_len:
        search_offset = data.find(search_key, search_offset, data_len)
        if search_offset == -1:
            break

        field_range_in_block = attributes.get_field_range_in_block(
            attributes.tail_block_format, search_field
        )
        if field_range_in_block is not None:
            tail_start_offset = search_offset - field_range_in_block.start
            tail_read_magic = attributes.get_field_data_in_block(
                attributes.tail_block_format, search_field, data, tail_start_offset
            )
            if tail_read_magic == search_key:
                tail_block_size = attributes.get_block_size(attributes.tail_block_format)
                tail_end_offset = tail_start_offset + tail_block_size
                return tail_end_offset

        search_offset += 1
    return data_len


def _get_end_from_size(
    attributes: FlashAttributes,
    start_index: int,
    data: bytes,
    data_len: int,
    field_type: FlashFieldType,
):
    # Only calculates the size if it knows the total size, or data size
    if (field_type is not FlashFieldType.TOTAL_SIZE) and (
        field_type is not FlashFieldType.DATA_SIZE
    ):
        return data_len

    if attributes.tail_block_format is None:
        raise UnpackerError("Tried to find end of resource without providing tail block format")
    tail_block_size = attributes.get_block_size(attributes.tail_block_format)

    cur_offset = start_index
    total_data_size = 0
    read_offset = 0
    for c in attributes.iterate_through_all_blocks(data_len - start_index, True):
        cur_block_size = attributes.get_block_size(c)

        # Treat every block as the tail, checking if it has the right field
        cur_block_size_field = attributes.get_field_data_in_block(
            attributes.tail_block_format, field_type, data, cur_offset
        )

        if cur_block_size_field is not None:
            read_offset = int.from_bytes(cur_block_size_field, "big")
            end_rel_offset = (cur_offset - start_index) + tail_block_size

            if (
                attributes.get_field_in_block(
                    attributes.tail_block_format, FlashFieldType.TOTAL_SIZE
                )
                is not None
                and read_offset == end_rel_offset
            ) or (
                attributes.get_field_in_block(
                    attributes.tail_block_format, FlashFieldType.DATA_SIZE
                )
                is not None
                and read_offset == total_data_size
            ):
                return cur_offset + tail_block_size
        total_data_size += attributes.get_field_length_in_block(c, FlashFieldType.DATA)
        cur_offset += cur_block_size
    return data_len


def _build_block(
    cur_block_type: Iterable[FlashField],
    attributes: FlashAttributes,
    block_data: bytes,
    original_data: bytes,
) -> bytes:
    # Update the checksum, even if its not used we use it for tracking if we need it to update ECC
    if attributes is None:
        raise UnpackerError("Cannot pack without providing FlashAttributes")
    data_hash = md5(block_data).digest()
    ecc_attr = attributes.ecc_attributes
    block = b""
    if ecc_attr is not None:
        ecc_class = ecc_attr.ecc_class
        if ecc_class is None:
            raise UnpackerError("Cannot pack FlashLogicalDataResource without providing ECC class")
        for field in cur_block_type:
            if field is not None:
                f = field.field_type
                f_size = field.size
                if f is FlashFieldType.ALIGNMENT:
                    block += b"\x00" * f_size
                elif f is FlashFieldType.CHECKSUM:
                    if attributes.checksum_func is not None:
                        block += attributes.checksum_func(original_data)
                elif f is FlashFieldType.DATA:
                    expected_data_size = attributes.get_field_length_in_block(
                        cur_block_type, FlashFieldType.DATA
                    )
                    real_data_len = len(block_data)
                    if real_data_len < expected_data_size:
                        cur_block_data = bytearray(expected_data_size)
                        cur_block_data[:real_data_len] = block_data
                        block_data = cur_block_data
                    block += block_data
                elif f is FlashFieldType.DATA_SIZE:
                    block += len(original_data).to_bytes(f_size, "big")
                elif f is FlashFieldType.DELIMITER:
                    if cur_block_type == attributes.header_block_format:
                        delimiter = ecc_attr.head_delimiter
                    elif cur_block_type == attributes.first_data_block_format:
                        delimiter = ecc_attr.first_data_delimiter
                    elif cur_block_type == attributes.data_block_format:
                        delimiter = ecc_attr.data_delimiter
                    elif cur_block_type == attributes.last_data_block_format:
                        delimiter = ecc_attr.last_data_delimiter
                    elif cur_block_type == attributes.tail_block_format:
                        delimiter = ecc_attr.tail_delimiter
                    else:
                        raise PackerError(
                            "Tried to add delimiter without specifying in FlashEccAttributes"
                        )
                    if delimiter is not None:
                        block += delimiter
                elif f is FlashFieldType.ECC:
                    if data_hash in DATA_HASHES:
                        ecc = DATA_HASHES[data_hash]
                    else:
                        # Assumes that all previously added data in the block should be included in the ECC
                        # TODO: Support ECC that comes before data
                        try:
                            if ecc_attr.ecc_class is not None:
                                ecc = ecc_attr.ecc_class.encode(block)
                        except TypeError:
                            raise PackerError(
                                "Tried to encode ECC without specifying ecc_class in FlashEccAttributes"
                            )
                    block += ecc
                elif f is FlashFieldType.ECC_SIZE:
                    block_ecc_field = attributes.get_field_in_block(
                        cur_block_type, FlashFieldType.ECC
                    )
                    if block_ecc_field is not None:
                        block += block_ecc_field.size.to_bytes(f_size, "big")
                elif f is FlashFieldType.TOTAL_SIZE:
                    data_size = len(original_data)
                    oob_size = attributes.get_total_oob_size(data_len=data_size)
                    expected_data_size = attributes.get_total_field_size(
                        data_len=data_size, field_type=FlashFieldType.DATA
                    )
                    total_size = expected_data_size + oob_size
                    block += (total_size).to_bytes(f_size, "big")
                elif f is FlashFieldType.MAGIC:
                    if ecc_attr.ecc_magic is None:
                        raise PackerError(
                            "Tried to add Magic without specifying in FlashEccAttributes"
                        )
                    block += ecc_attr.ecc_magic
    return block



import sys

from ofrak.cli.command.deps import DepsCommand
from ofrak.cli.command.gui import GUICommand
from ofrak.cli.command.identify import IdentifyCommand
from ofrak.cli.command.list import ListCommand
from ofrak.cli.command.unpack import UnpackCommand
from ofrak.cli.ofrak_cli import OFRAKCommandLineInterface


def main():  # pragma: no cover
    ofrak_cli = OFRAKCommandLineInterface(
        (ListCommand(), DepsCommand(), IdentifyCommand(), UnpackCommand(), GUICommand())
    )
    ofrak_cli.parse_and_run(sys.argv[1:])


if __name__ == "__main__":
    main()



import argparse

from ofrak import OFRAKContext
from ofrak.cli.ofrak_cli import OfrakCommandRunsScript
from ofrak.core.magic import Magic
from ofrak.gui.server import open_gui
from ofrak.resource import Resource


class IdentifyCommand(OfrakCommandRunsScript):
    def create_parser(self, parser: argparse._SubParsersAction):
        subparser = parser.add_parser(
            "identify",
            help="Identify all known structures in the binary",
            description="Import a file as an OFRAK resource, then identifies it and prints all the"
            " tags and attributes.",
        )
        subparser.add_argument("filename", help="File to identify")
        self.add_ofrak_arguments(subparser)

        # GUI args
        subparser.add_argument(
            "--gui",
            action="store_true",
            help="Open the OFRAK GUI after unpacking",
            default=False,
        )
        subparser.add_argument(
            "-gH",
            "--gui-hostname",
            action="store",
            help="Set GUI server host address.",
            default="127.0.0.1",
        )
        subparser.add_argument(
            "-gp",
            "--gui-port",
            action="store",
            type=int,
            help="Set GUI server host port.",
            default=8080,
        )
        subparser.add_argument(
            "--gui-no-browser",
            action="store_true",
            help="Don't open the browser to the OFRAK GUI",
        )

        return subparser

    async def ofrak_func(self, ofrak_context: OFRAKContext, args: argparse.Namespace):
        print(f"Identifying file: {args.filename}\n")
        root_resource = await ofrak_context.create_root_resource_from_file(args.filename)

        await root_resource.identify()
        print(await IdentifyCommand.print_info(root_resource))

        if args.gui:
            server = await open_gui(
                args.gui_hostname,
                args.gui_port,
                focus_resource=root_resource,
                open_in_browser=(not args.gui_no_browser),
            )
            await server.run_until_cancelled()

    @staticmethod
    async def print_info(resource: Resource) -> str:
        output = ""

        output += "= Tags = \n"
        for tag in resource.get_tags():
            output += f"  {tag}\n"

        output += "= Attributes =\n"
        for attributes_t, attribute in resource.get_model().attributes.items():
            if attributes_t is Magic:
                output += (
                    f"  Magic: "
                    f"\n   Mime: {attribute.mime}"
                    f"\n   Descriptor: {attribute.descriptor}\n"
                )
            else:
                output += f"  {attribute}"

        return output



import asyncio
import logging
import os
import tempfile
from dataclasses import dataclass
from subprocess import CalledProcessError

from ofrak.component.packer import Packer
from ofrak.component.unpacker import Unpacker
from ofrak.resource import Resource
from ofrak.core.filesystem import File, Folder, FilesystemRoot, SpecialFileType
from ofrak.core.magic import MagicMimeIdentifier, MagicDescriptionIdentifier
from ofrak.core.binary import GenericBinary

from ofrak.model.component_model import ComponentExternalTool
from ofrak_type.range import Range

LOGGER = logging.getLogger(__name__)

ZIP_TOOL = ComponentExternalTool(
    "zip",
    "https://linux.die.net/man/1/zip",
    install_check_arg="--help",
    apt_package="zip",
    brew_package="zip",
)
UNZIP_TOOL = ComponentExternalTool(
    "unzip",
    "https://linux.die.net/man/1/unzip",
    install_check_arg="--help",
    apt_package="unzip",
    brew_package="unzip",
)


@dataclass
class ZipArchive(GenericBinary, FilesystemRoot):
    """
    Filesystem stored in a zip archive.
    """


class ZipUnpacker(Unpacker[None]):
    """
    Unpack (decompress) a zip archive.
    """

    targets = (ZipArchive,)
    children = (File, Folder, SpecialFileType)
    external_dependencies = (UNZIP_TOOL,)

    async def unpack(self, resource: Resource, config=None):
        zip_view = await resource.view_as(ZipArchive)
        with tempfile.NamedTemporaryFile(suffix=".zip") as temp_archive:
            temp_archive.write(await resource.get_data())
            temp_archive.flush()
            with tempfile.TemporaryDirectory() as temp_dir:
                cmd = [
                    "unzip",
                    temp_archive.name,
                    "-d",
                    temp_dir,
                ]
                proc = await asyncio.create_subprocess_exec(
                    *cmd,
                )
                returncode = await proc.wait()
                if proc.returncode:
                    raise CalledProcessError(returncode=returncode, cmd=cmd)
                await zip_view.initialize_from_disk(temp_dir)


class ZipPacker(Packer[None]):
    """
    Pack files into a compressed zip archive.
    """

    targets = (ZipArchive,)
    external_dependencies = (ZIP_TOOL,)

    async def pack(self, resource: Resource, config=None):
        zip_view: ZipArchive = await resource.view_as(ZipArchive)
        flush_dir = await zip_view.flush_to_disk()
        temp_archive = f"{flush_dir}.zip"
        cwd = os.getcwd()
        os.chdir(flush_dir)
        cmd = [
            "zip",
            "-r",
            temp_archive,
            ".",
        ]
        proc = await asyncio.create_subprocess_exec(
            *cmd,
        )
        returncode = await proc.wait()
        if proc.returncode:
            raise CalledProcessError(returncode=returncode, cmd=cmd)
        os.chdir(cwd)
        with open(temp_archive, "rb") as fh:
            resource.queue_patch(Range(0, await zip_view.resource.get_data_length()), fh.read())


MagicMimeIdentifier.register(ZipArchive, "application/zip")
MagicDescriptionIdentifier.register(
    ZipArchive, lambda desc: any([("Zip archive data" in s) for s in desc.split(", ")])
)



from typing import Any, Tuple, Union, List, Callable

from beartype import beartype
from typing_inspect import get_origin, get_args

from ofrak.service.serialization.serializers.serializer_i import SerializerInterface


class TupleSerializer(SerializerInterface):
    """
    Serialize and deserialize `Tuple[...]` into `PJSONType`.
    """

    targets = (lambda type_hint: get_origin(type_hint) == tuple,)

    @beartype
    def obj_to_pjson(self, obj: Tuple, type_hint: Any) -> Tuple:
        return self._handle(obj, type_hint, self._service.to_pjson)

    @beartype
    def pjson_to_obj(self, pjson_obj: Union[List, Tuple], type_hint: Any) -> Tuple:
        return self._handle(pjson_obj, type_hint, self._service.from_pjson)

    def _handle(self, obj: Union[List, Tuple], type_hint: Any, factory: Callable) -> Tuple:
        args = get_args(type_hint)
        if Ellipsis in args:
            # Tuple[X, ...]. All items are expected to be of type `X`.
            return tuple(factory(item, args[0]) for item in obj)
        else:
            # e.g. Tuple[X, Y, Z]
            if len(args) != len(obj):
                raise TypeError(f"invalid tuple size for {obj} with expected type {type_hint}")
            return tuple(factory(item, arg) for item, arg in zip(obj, args))



import os
import stat
import tempfile
from dataclasses import dataclass
from typing import Dict, Iterable, Optional, Type, Union

try:
    import xattr
except ImportError:
    import ofrak.core.xattr_stub as xattr  # type: ignore[no-redef]

from ofrak.model.viewable_tag_model import AttributesType
from ofrak.resource import Resource

from ofrak.model.resource_model import index, ResourceAttributes
from ofrak.model.tag_model import ResourceTag
from ofrak.resource_view import ResourceView
from ofrak.service.resource_service_i import (
    ResourceFilter,
    ResourceAttributeValueFilter,
    ResourceFilterCondition,
)


@dataclass
class FilesystemEntry(ResourceView):
    """
    Handles generic management of any entry stored within a filesystem.
    """

    name: str
    stat: Optional[os.stat_result]
    xattrs: Optional[Dict[str, bytes]]

    @index
    def Name(self) -> str:
        name = self.name.rstrip("/")
        return name.split("/")[-1]

    @classmethod
    def caption(cls, all_attributes) -> str:
        try:
            filesystem_attributes = all_attributes[AttributesType[FilesystemEntry]]
        except KeyError:
            return super().caption(all_attributes)
        return f"{cls.__name__}: {filesystem_attributes.name}"

    def get_name(self) -> str:
        """
        Get the base name of a folder.

        :return: The file or folder's base name
        """
        return self.Name

    async def set_stat(self, stat_result: os.stat_result):
        """
        Set the stat for the `FilesystemEntry`. Useful for newly created files where we want to
        control the full 10-tuple.

        :param stat_result: `os.stat_result` object containing all necessary values
        """
        self.stat = stat_result
        if self.resource is None:
            return
        all_view_attrs: Dict[
            Type[ResourceAttributes], ResourceAttributes
        ] = self.get_attributes_instances()
        filesystem_attrs = all_view_attrs[AttributesType[FilesystemEntry]]
        self.resource.add_attributes(filesystem_attrs)
        await self.resource.save()

    async def modify_stat_attribute(self, st_stat: int, stat_value: int):
        """
        Modify a specific `os.stat` attribute on the filesystem entry.

        Example:

        ```python
        fs_entry.modify_stat_attribute(stat.ST_MODE, 0o100755)
        ```

        :param st_stat: The `st_stat` struct member to be modified
        :param stat_value: The new stat value
        """
        if self.stat is None:
            raise ValueError("Cannot modify a stat attribute when stat attributes are not set")
        stat_attributes = list(self.stat)  # type: ignore
        stat_attributes[st_stat] = stat_value
        filesystem_stat_attributes = os.stat_result(tuple(stat_attributes))
        await self.set_stat(filesystem_stat_attributes)

    async def modify_xattr_attribute(self, attribute: str, value: bytes):
        """
        Modify the extended file attributes ("xattrs") for the filesystem entry.

        :param attribute:
        :param value:
        """
        if self.xattrs is None:
            self.xattrs = dict()
        self.xattrs[attribute] = value
        if self.resource is None:
            return
        await self.resource.save()

    async def get_path(self) -> str:
        """
        Get a folder's path, with the `FilesystemRoot` as the path root.

        :return: The full path name, with the `FilesystemRoot` ancestor as the path root
        """
        path = [self.get_name()]

        for a in await self.resource.get_ancestors(
            r_filter=ResourceFilter(
                tags=(FilesystemEntry, FilesystemRoot),
                tags_condition=ResourceFilterCondition.OR,
            )
        ):
            if (a is None) or (a.has_tag(FilesystemRoot)):
                break
            a_view = await a.view_as(FilesystemEntry)
            path.append(a_view.get_name())
        path.reverse()

        return os.path.join(*path)

    def apply_stat_attrs(self, path: str):
        """
        Set file mode and access times of a path on disk to match the attributes stored on this
        resource.

        :param path: Path on disk to set attributes of.
        """
        if self.stat:
            os.chown(path, self.stat.st_uid, self.stat.st_gid)
            os.chmod(path, self.stat.st_mode)
            os.utime(path, (self.stat.st_atime, self.stat.st_mtime))
        if self.xattrs:
            for attr, value in self.xattrs.items():
                xattr.setxattr(path, attr, value)

    def is_file(self) -> bool:
        return self.resource.has_tag(File)

    def is_folder(self) -> bool:
        return self.resource.has_tag(Folder)

    def is_link(self) -> bool:
        return self.resource.has_tag(SymbolicLink)

    def is_block_device(self) -> bool:
        return self.resource.has_tag(BlockDevice)

    def is_fifo_pipe(self) -> bool:
        return self.resource.has_tag(FIFOPipe)

    def is_character_device(self) -> bool:
        return self.resource.has_tag(CharacterDevice)

    def is_device(self) -> bool:
        return self.is_block_device() or self.is_character_device()


class File(FilesystemEntry):
    """
    Stores the data and location of a file within a filesystem or folder's descendant file tree.
    """


class Folder(FilesystemEntry):
    """
    Describes a folder that is stored in a filesystem as a file tree.

    All descendant resources are stored in a tree structure that reflects a folder/directory's file
    tree.
    """

    async def get_entry(self, path: str) -> Optional[FilesystemEntry]:
        """
        Search a folder for an entry with the given path.

        :param path: The filesystem path to search for, relative to this folder

        :return: The child `FilesystemEntry` resource that was found. If nothing was found, `None`
        is returned
        """
        basename = os.path.basename(path)
        # only searching paths with the same base name should reduce the search space by quite a lot
        descendants = await self.resource.get_descendants_as_view(
            FilesystemEntry,
            r_filter=ResourceFilter(
                attribute_filters=(ResourceAttributeValueFilter(FilesystemEntry.Name, basename),)
            ),
        )

        for d in descendants:
            descendant_path = await d.get_path()
            if descendant_path.split(f"{self.name}/")[-1] == path:
                return d
        return None

    async def list_dir(self) -> Dict[str, FilesystemEntry]:
        """
        Enumerate a folder's children, much like `os.listdir`.

        :return: A dictionary of child entries, with the child's name as the key
        """
        entries = dict()
        for c in await self.resource.get_children_as_view(FilesystemEntry):
            entries[c.get_name()] = c
        return entries


@dataclass
class SpecialFileType(FilesystemEntry):
    """
    A filesystem entry that is not a simple type, like a file or folder. For example, symbolic
    links and block devices fall under this category.
    """


@dataclass
class SymbolicLink(SpecialFileType):
    """
    Symbolic link pointing to a (possibly invalid) path in the filesystem. The path pointed to is
    invalid if accessing it would raise a `FileNotFoundError`.

    :ivar source_path: File pointed to by the symbolic link
    """

    source_path: str


@dataclass
class BlockDevice(SpecialFileType):
    """
    Block special device file.
    """


@dataclass
class FIFOPipe(SpecialFileType):
    """
    Named pipe.
    """


@dataclass
class CharacterDevice(SpecialFileType):
    """
    Character special device file.
    """


@dataclass
class FilesystemRoot(ResourceView):
    """
    A resource that contains a filesystem's file tree. All descendant resources are stored in a tree
    structure that reflects a filesystem's file tree. The methods within this class are intended to
    be used as utilities when unpacking any filesystem-like resource.

    Any resource that contains a file tree should inherit the `FilesystemRoot` class.
    """

    async def initialize_from_disk(
        self,
        path: str,
    ):
        root_path = os.path.normpath(path)

        if not os.path.exists(root_path):
            raise FileNotFoundError(
                f"Could not initialize from disk. Root path does not exist: {root_path}"
            )
        if not os.path.isdir(root_path):
            raise FileNotFoundError(
                f"Could not initialize from disk. Found a file instead of a directory: {root_path}"
            )

        for root, dirs, files in os.walk(root_path):
            for d in sorted(dirs):
                absolute_path = os.path.join(root, d)
                relative_path = os.path.join(os.path.relpath(root, root_path), d)
                folder_attributes_stat = os.lstat(absolute_path)

                mode = folder_attributes_stat.st_mode
                mode_tests = [
                    stat.S_ISCHR,
                    stat.S_ISBLK,
                    stat.S_ISFIFO,
                    stat.S_ISSOCK,
                    stat.S_ISDOOR,
                    stat.S_ISPORT,
                    stat.S_ISWHT,
                    stat.S_ISREG,
                ]
                for mode_test in mode_tests:
                    if mode_test(mode) != 0:
                        raise NotImplementedError(
                            f"Directory {absolute_path} has an unsupported special file type: "
                            f"{stat.S_IFMT(mode):o}. {mode_test.__name__} should be false."
                        )

                folder_attributes_xattr = self._get_xattr_map(absolute_path)
                if os.path.islink(absolute_path):
                    await self.add_special_file_entry(
                        relative_path,
                        SymbolicLink(
                            relative_path,
                            folder_attributes_stat,
                            folder_attributes_xattr,
                            os.readlink(absolute_path),
                        ),
                    )
                else:
                    await self.add_folder(
                        relative_path,
                        folder_attributes_stat,
                        folder_attributes_xattr,
                    )
            for f in sorted(files):
                absolute_path = os.path.join(root, f)
                relative_path = os.path.normpath(os.path.join(os.path.relpath(root, root_path), f))
                file_attributes_stat = os.lstat(absolute_path)

                mode = file_attributes_stat.st_mode
                mode_tests = [
                    stat.S_ISSOCK,
                    stat.S_ISDOOR,
                    stat.S_ISPORT,
                    stat.S_ISWHT,
                    stat.S_ISDIR,
                ]
                for mode_test in mode_tests:
                    if mode_test(mode) != 0:
                        raise NotImplementedError(
                            f"Directory {absolute_path} has an unsupported special file type: "
                            f"{stat.S_IFMT(mode):o}. {mode_test.__name__} should be false."
                        )

                file_attributes_xattr = self._get_xattr_map(absolute_path)
                if os.path.islink(absolute_path):
                    await self.add_special_file_entry(
                        relative_path,
                        SymbolicLink(
                            relative_path,
                            file_attributes_stat,
                            file_attributes_xattr,
                            os.readlink(absolute_path),
                        ),
                    )
                elif os.path.isfile(absolute_path):
                    with open(absolute_path, "rb") as fh:
                        await self.add_file(
                            relative_path,
                            fh.read(),
                            file_attributes_stat,
                            file_attributes_xattr,
                        )
                elif stat.S_ISFIFO(mode):
                    await self.add_special_file_entry(
                        relative_path,
                        FIFOPipe(relative_path, file_attributes_stat, file_attributes_xattr),
                    )
                elif stat.S_ISBLK(mode):
                    await self.add_special_file_entry(
                        relative_path,
                        BlockDevice(relative_path, file_attributes_stat, file_attributes_xattr),
                    )
                elif stat.S_ISCHR(mode):
                    await self.add_special_file_entry(
                        relative_path,
                        CharacterDevice(relative_path, file_attributes_stat, file_attributes_xattr),
                    )
                else:
                    raise NotImplementedError(
                        f"File {absolute_path} appeared to be a supported "
                        f"type but did not match any of the known cases to "
                        f"create a resource. Stat: {stat.S_IFMT(mode):o}"
                    )

    async def flush_to_disk(
        self,
        path: Optional[str] = None,
    ):
        """
        Writes this `FilesystemRoot`'s `FilesystemEntry` descendants to directory. If a target path
        is not provided, the output is written to a temporary directory.

        :return: the root directory containing the flushed filesystem
        """
        if path is None:
            root_path = tempfile.mkdtemp()
        else:
            root_path = path

        entries = [
            f
            for f in await self.resource.get_children_as_view(
                FilesystemEntry, r_filter=ResourceFilter(tags=(FilesystemEntry,))
            )
        ]
        while len(entries) > 0:
            entry = entries.pop(0)
            entry_path = await entry.get_path()
            if entry.is_link():
                link_name = os.path.join(root_path, entry_path)
                if not os.path.exists(link_name):
                    link_view = await entry.resource.view_as(SymbolicLink)
                    os.symlink(link_view.source_path, link_name)
                assert len(list(await entry.resource.get_children())) == 0
                if entry.stat:
                    # https://docs.python.org/3/library/os.html#os.supports_follow_symlinks
                    if os.chown in os.supports_follow_symlinks:
                        os.chown(
                            link_name, entry.stat.st_uid, entry.stat.st_gid, follow_symlinks=False
                        )
                    if os.chmod in os.supports_follow_symlinks:
                        os.chmod(link_name, entry.stat.st_mode, follow_symlinks=False)
                    if os.utime in os.supports_follow_symlinks:
                        os.utime(
                            link_name,
                            (entry.stat.st_atime, entry.stat.st_mtime),
                            follow_symlinks=False,
                        )
                if entry.xattrs:
                    for attr, value in entry.xattrs.items():
                        xattr.setxattr(link_name, attr, value, symlink=True)  # Don't follow links
            elif entry.is_folder():
                folder_name = os.path.join(root_path, entry_path)
                if not os.path.exists(folder_name):
                    os.makedirs(folder_name)
                for child in await entry.resource.get_children_as_view(
                    FilesystemEntry, r_filter=ResourceFilter(tags=(FilesystemEntry,))
                ):
                    entries.append(child)
            elif entry.is_file():
                file_name = os.path.join(root_path, entry_path)
                with open(file_name, "wb") as f:
                    f.write(await entry.resource.get_data())
                entry.apply_stat_attrs(file_name)
            elif entry.is_device():
                device_name = os.path.join(root_path, entry_path)
                if entry.stat is None:
                    raise ValueError(
                        f"Cannot create a device {entry_path} for a "
                        f"BlockDevice or CharacterDevice resource with no stat!"
                    )
                os.mknod(device_name, entry.stat.st_mode, entry.stat.st_rdev)
                entry.apply_stat_attrs(device_name)
            elif entry.is_fifo_pipe():
                fifo_name = os.path.join(root_path, entry_path)
                if entry.stat is None:
                    raise ValueError(
                        f"Cannot create a fifo {entry_path} for a FIFOPipe resource "
                        "with no stat!"
                    )
                os.mkfifo(fifo_name, entry.stat.st_mode)
                entry.apply_stat_attrs(fifo_name)
            else:
                entry_info = f"Stat: {stat.S_IFMT(entry.stat.st_mode):o}" if entry.stat else ""
                raise NotImplementedError(
                    f"FilesystemEntry {entry_path} has an unknown or "
                    f"unsupported filesystem type! Unable to create it "
                    f"on-disk. {entry_info}"
                )

        return root_path

    async def get_entry(self, path: str):
        """
        Searches this `FilesystemRoot`'s descendants for a filesystem entry with a given path,
            and returns that entry if found.

        :param path: the path of the `FilesystemEntry` to search for, with this `FilesystemRoot` as
        the path root

        :return: the descendant `FilesystemEntry`, if found; otherwise, returns `None`
        """
        basename = os.path.basename(path)
        # only searching paths with the same base name should reduce the search space by quite a lot
        descendants = await self.resource.get_descendants_as_view(
            FilesystemEntry,
            r_filter=ResourceFilter(
                attribute_filters=(ResourceAttributeValueFilter(FilesystemEntry.Name, basename),)
            ),
        )

        for d in descendants:
            if await d.get_path() == os.path.normpath(path):
                return d
        return None

    async def list_dir(self) -> dict:
        """
        Enumerates a `FilesystemRoot`'s children, much like `os.listdir`.

        :return: a dictionary of child entries, with the child's name as the key
        """
        entries = dict()
        for c in await self.resource.get_children_as_view(FilesystemEntry):
            entries[c.get_name()] = c
        return entries

    async def add_folder(
        self,
        path: str,
        folder_stat_result: Optional[os.stat_result] = None,
        folder_xattrs: Optional[Dict[str, bytes]] = None,
        tags: Iterable[ResourceTag] = (),
        attributes: Iterable[ResourceAttributes] = (),
    ) -> Folder:
        """
        Adds a [Folder][ofrak.core.filesystem.Folder] resource to a `FilesystemRoot`, creating all
        parent folders as needed.

        :param path: the path that will contain the folder to be added
        :param folder_stat_result: the filesystem attributes associated with the folder
        :param folder_xattrs: xattrs for the folder
        :param tags: the list of tags to be added to the new resource. The `Folder` tag is added by
        default
        :param attributes: the list of additional attributes to be added to the new folder, the
        folder's name attribute is added automatically

        :raises ValueError: if the path is too short and doesn't actually include any directories

        :return: the `Folder` resource that was added to the `FilesystemRoot`
        """
        # Normalizes and cleans up paths beginning with "./" and containing "./../" as well as
        # other extraneous separators
        split_dir = os.path.normpath(path).rstrip("/").lstrip("/").split("/")

        parent: Union[FilesystemRoot, Folder] = self
        for directory in split_dir:
            folder_entries = await parent.list_dir()

            if directory not in folder_entries.keys():
                new_missing_folder = await parent.resource.create_child_from_view(
                    Folder(directory, folder_stat_result, folder_xattrs),
                    data=b"",
                    additional_tags=tags,
                    additional_attributes=attributes,
                )
                parent = await new_missing_folder.view_as(Folder)
            else:
                parent = await folder_entries[directory].resource.view_as(Folder)

        if type(parent) is FilesystemRoot:
            assert len(split_dir) == 0  # Only case this should happen
            raise ValueError(f"The path {path} is too short (no directories)")

        if not isinstance(parent, Folder):
            raise ValueError(
                f"Parent folder {parent} has an unexpected type {type(parent)}. It "
                f"should be a Folder instead."
            )
        return parent

    async def add_file(
        self,
        path: str,
        data: bytes,
        file_stat_result: Optional[os.stat_result] = None,
        file_xattrs: Optional[Dict[str, bytes]] = None,
        tags: Iterable[ResourceTag] = (),
        attributes: Iterable[ResourceAttributes] = (),
    ) -> Resource:
        """
        Adds a [File][ofrak.core.filesystem.File] resource to a `FilesystemRoot`, creating all
        parent [Folders][ofrak.core.filesystem.Folder] as needed.

        :param path: the path that will contain the `File` to be added
        :param data: contents of the file being added
        :param file_stat_result: the filesystem attributes associated with the file
        :param file_xattrs: xattrs for the file
        :param tags: the list of tags to be added to the new resource, the File tag is added by
        default
        :param attributes: the list of additional attributes to be added to the new Folder,
            the file's name attribute is added automatically

        :return: the `File` resource that was added to the `FilesystemRoot`
        """
        dirname = os.path.dirname(path)
        filename = os.path.basename(path)

        if dirname == "":
            parent_folder = self
        else:
            parent_folder = await self.get_entry(dirname)
            if parent_folder is None:
                parent_folder = await self.add_folder(dirname)

        new_file = await parent_folder.resource.create_child_from_view(
            File(filename, file_stat_result, file_xattrs),
            data=data,
            additional_tags=tags,
            additional_attributes=attributes,
        )
        return new_file

    async def remove_file(self, path: str) -> None:
        """
        Removes a [File][ofrak.core.filesystem.File] resource from a `FilesystemRoot`.
        :param path: the path of the file to be removed
        :return: None
        """
        file_to_remove = await self.get_entry(path)
        await file_to_remove.resource.delete()
        await file_to_remove.resource.save()

    async def add_special_file_entry(
        self,
        path: str,
        special_file_view: SpecialFileType,
        tags: Iterable[ResourceTag] = (),
        attributes: Iterable[ResourceAttributes] = (),
    ) -> Resource:
        """
        Adds a resource representing a [SpecialFileType][ofrak.core.filesystem.SpecialFileType]
        to a `FilesystemRoot`, creating all parent [Folders][ofrak.core.filesystem.Folder] as
        needed.

        Some examples of these "special" types are
        [SymbolicLink][ofrak.core.filesystem.SymbolicLink] and
        [BlockDevice][ofrak.core.filesystem.BlockDevice].

        :param path: The path of the `FilesystemEntry` to be added
        :param special_file_view: A ResourceView, whose type should be a subclass of
        `FilesystemEntry`
        :param tags: the list of tags to be added to the new resource, the File tag is added by
        default
        :param attributes: the list of additional attributes to be added to the new Folder,
        the file's name attribute is added automatically

        :return: The special `FilesystemEntry` resource that was added to the `FilesystemRoot`
        """
        dirname = os.path.dirname(path)

        if dirname == "":
            parent_folder = self
        else:
            parent_folder = await self.get_entry(dirname)
            if parent_folder is None:
                parent_folder = await self.add_folder(dirname)

        new_entry = await parent_folder.resource.create_child_from_view(
            special_file_view,
            data=b"",  # Use empty, non-None data
            additional_tags=tags,
            additional_attributes=attributes,
        )
        return new_entry

    @classmethod
    def _get_xattr_map(cls, path):
        xattr_dict = {}
        for attr in xattr.listxattr(path, symlink=True):  # Don't follow links
            xattr_dict[attr] = xattr.getxattr(path, attr)
        return xattr_dict



from dataclasses import dataclass
from typing import Tuple

import pytest

from ofrak import OFRAKContext
from ofrak_type.architecture import InstructionSetMode
from ofrak.core.architecture import ProgramAttributes
from ofrak.core.instruction import Instruction, RegisterUsage
from ofrak.service.assembler.assembler_service_i import AssemblerServiceInterface
from test_ofrak.constants import ARM32_ARCH, X64_ARCH, PPC_ARCH


@dataclass
class RegisterAnalyzerTestCase:
    program_attributes: ProgramAttributes
    instruction: Instruction
    expected_regs_read: Tuple[str, ...]
    expected_regs_written: Tuple[str, ...]

    @property
    def label(self):
        return f"{self.program_attributes.isa.name}: {self.instruction.disassembly}"


ARM_REGISTER_USAGE_TEST_CASES = [
    RegisterAnalyzerTestCase(
        ARM32_ARCH,
        Instruction(
            0x100,
            0x4,
            "push {r4, r5, r6, r7, r8, lr}",
            "push",
            "{r4, r5, r6, r7, r8, lr}",
            InstructionSetMode.NONE,
        ),
        ("sp", "r4", "r5", "r6", "r7", "r8", "lr"),
        ("sp",),
    ),
    RegisterAnalyzerTestCase(
        ARM32_ARCH,
        Instruction(
            0x104,
            0x4,
            "mov r4, #0x0",
            "mov",
            "r4, #0x0",
            InstructionSetMode.NONE,
        ),
        (),
        ("r4",),
    ),
    RegisterAnalyzerTestCase(
        ARM32_ARCH,
        Instruction(
            0x108,
            0x4,
            "mov r5, r4",
            "mov",
            "r5, r4",
            InstructionSetMode.NONE,
        ),
        ("r4",),
        ("r5",),
    ),
    RegisterAnalyzerTestCase(
        ARM32_ARCH,
        Instruction(
            0x10C,
            0x4,
            "ldr r6, [pc, #0x50]",
            "ldr",
            "r6, [pc, #0x50]",
            InstructionSetMode.NONE,
        ),
        ("pc",),
        ("r6",),
    ),
    RegisterAnalyzerTestCase(
        ARM32_ARCH,
        Instruction(
            0x110,
            0x4,
            "ldr r7, [pc, #0x50]",
            "ldr",
            "r7, [pc, #0x50]",
            InstructionSetMode.NONE,
        ),
        ("pc",),
        ("r7",),
    ),
]


X64_REGISTER_USAGE_TEST_CASES = [
    RegisterAnalyzerTestCase(
        X64_ARCH,
        Instruction(
            virtual_address=0x4003E0,
            size=2,
            disassembly="xor ebp, ebp",
            mnemonic="xor",
            operands="ebp, ebp",
            mode=InstructionSetMode.NONE,
        ),
        ("ebp",),
        ("rflags", "ebp"),
    ),
    RegisterAnalyzerTestCase(
        X64_ARCH,
        Instruction(
            virtual_address=0x4003E2,
            size=3,
            disassembly="mov r9, rdx",
            mnemonic="mov",
            operands="r9, rdx",
            mode=InstructionSetMode.NONE,
        ),
        ("rdx",),
        ("r9",),
    ),
    RegisterAnalyzerTestCase(
        X64_ARCH,
        Instruction(
            virtual_address=0x4003E5,
            size=1,
            disassembly="pop rsi",
            mnemonic="pop",
            operands="rsi",
            mode=InstructionSetMode.NONE,
        ),
        ("rsp",),
        ("rsp", "rsi"),
    ),
    RegisterAnalyzerTestCase(
        X64_ARCH,
        Instruction(
            virtual_address=0x4003E6,
            size=3,
            disassembly="mov rdx, rsp",
            mnemonic="mov",
            operands="rdx, rsp",
            mode=InstructionSetMode.NONE,
        ),
        ("rsp",),
        ("rdx",),
    ),
    RegisterAnalyzerTestCase(
        X64_ARCH,
        Instruction(
            virtual_address=0x4003ED,
            size=1,
            disassembly="push rax",
            mnemonic="push",
            operands="rax",
            mode=InstructionSetMode.NONE,
        ),
        ("rsp", "rax"),
        ("rsp",),
    ),
    RegisterAnalyzerTestCase(
        X64_ARCH,
        Instruction(
            virtual_address=0x4003EE,
            size=1,
            disassembly="push rsp",
            mnemonic="push",
            operands="rsp",
            mode=InstructionSetMode.NONE,
        ),
        ("rsp",),
        ("rsp",),
    ),
    RegisterAnalyzerTestCase(
        X64_ARCH,
        Instruction(
            virtual_address=0x4003EF,
            size=7,
            disassembly="mov r8, 0x4004e0",
            mnemonic="mov",
            operands="r8, 0x4004e0",
            mode=InstructionSetMode.NONE,
        ),
        (),
        ("r8",),
    ),
    RegisterAnalyzerTestCase(
        X64_ARCH,
        Instruction(
            virtual_address=0x4003F6,
            size=7,
            disassembly="mov rcx, 0x4004f0",
            mnemonic="mov",
            operands="rcx, 0x4004f0",
            mode=InstructionSetMode.NONE,
        ),
        (),
        ("rcx",),
    ),
    RegisterAnalyzerTestCase(
        X64_ARCH,
        Instruction(
            virtual_address=0x4003FD,
            size=7,
            disassembly="mov rdi, 0x4004c4",
            mnemonic="mov",
            operands="rdi, 0x4004c4",
            mode=InstructionSetMode.NONE,
        ),
        (),
        ("rdi",),
    ),
    RegisterAnalyzerTestCase(
        X64_ARCH,
        Instruction(
            virtual_address=0x400404,
            size=5,
            disassembly="call 0x4003c8",
            mnemonic="call",
            operands="0x4003c8",
            mode=InstructionSetMode.NONE,
        ),
        ("rsp", "rip"),
        ("rsp",),
    ),
    RegisterAnalyzerTestCase(
        X64_ARCH,
        Instruction(
            virtual_address=0x400409,
            size=1,
            disassembly="hlt ",
            mnemonic="hlt",
            operands="",
            mode=InstructionSetMode.NONE,
        ),
        (),
        (),
    ),
    RegisterAnalyzerTestCase(
        X64_ARCH,
        Instruction(
            virtual_address=0x40040C,
            size=4,
            disassembly="sub rsp, 0x8",
            mnemonic="sub",
            operands="rsp, 0x8",
            mode=InstructionSetMode.NONE,
        ),
        ("rsp",),
        ("rflags", "rsp"),
    ),
    RegisterAnalyzerTestCase(
        X64_ARCH,
        Instruction(
            virtual_address=0x400410,
            size=7,
            disassembly="mov rax, qword ptr [0x600840]",
            mnemonic="mov",
            operands="rax, qword ptr [0x600840]",
            mode=InstructionSetMode.NONE,
        ),
        ("rip",),
        ("rax",),
    ),
    RegisterAnalyzerTestCase(
        X64_ARCH,
        Instruction(
            virtual_address=0x400417,
            size=3,
            disassembly="test rax, rax",
            mnemonic="test",
            operands="rax, rax",
            mode=InstructionSetMode.NONE,
        ),
        ("rax",),
        ("rflags",),
    ),
    RegisterAnalyzerTestCase(
        X64_ARCH,
        Instruction(
            virtual_address=0x40041A,
            size=2,
            disassembly="jz 0x40041e",
            mnemonic="jz",
            operands="0x40041e",
            mode=InstructionSetMode.NONE,
        ),
        ("rflags",),
        (),
    ),
    RegisterAnalyzerTestCase(
        X64_ARCH,
        Instruction(
            virtual_address=0x40041E,
            size=4,
            disassembly="add rsp, 0x8",
            mnemonic="add",
            operands="rsp, 0x8",
            mode=InstructionSetMode.NONE,
        ),
        ("rsp",),
        ("rflags", "rsp"),
    ),
    RegisterAnalyzerTestCase(
        X64_ARCH,
        Instruction(
            virtual_address=0x400422,
            size=1,
            disassembly="ret ",
            mnemonic="ret",
            operands="",
            mode=InstructionSetMode.NONE,
        ),
        ("rsp",),
        ("rsp",),
    ),
    RegisterAnalyzerTestCase(
        X64_ARCH,
        Instruction(
            virtual_address=0x40041C,
            size=2,
            disassembly="call rax",
            mnemonic="call",
            operands="rax",
            mode=InstructionSetMode.NONE,
        ),
        ("rsp", "rax"),
        ("rsp",),
    ),
    RegisterAnalyzerTestCase(
        X64_ARCH,
        Instruction(
            virtual_address=0x400430,
            size=1,
            disassembly="push rbp",
            mnemonic="push",
            operands="rbp",
            mode=InstructionSetMode.NONE,
        ),
        ("rsp", "rbp"),
        ("rsp",),
    ),
    RegisterAnalyzerTestCase(
        X64_ARCH,
        Instruction(
            virtual_address=0x400431,
            size=3,
            disassembly="mov rbp, rsp",
            mnemonic="mov",
            operands="rbp, rsp",
            mode=InstructionSetMode.NONE,
        ),
        ("rsp",),
        ("rbp",),
    ),
    RegisterAnalyzerTestCase(
        X64_ARCH,
        Instruction(
            virtual_address=0x400434,
            size=1,
            disassembly="push rbx",
            mnemonic="push",
            operands="rbx",
            mode=InstructionSetMode.NONE,
        ),
        ("rsp", "rbx"),
        ("rsp",),
    ),
    RegisterAnalyzerTestCase(
        X64_ARCH,
        Instruction(
            virtual_address=0x400435,
            size=4,
            disassembly="sub rsp, 0x8",
            mnemonic="sub",
            operands="rsp, 0x8",
            mode=InstructionSetMode.NONE,
        ),
        ("rsp",),
        ("rflags", "rsp"),
    ),
    RegisterAnalyzerTestCase(
        X64_ARCH,
        Instruction(
            virtual_address=0x400439,
            size=7,
            disassembly="cmp byte ptr [0x600880], 0x0",
            mnemonic="cmp",
            operands="byte ptr [0x600880], 0x0",
            mode=InstructionSetMode.NONE,
        ),
        ("rip",),
        ("rflags",),
    ),
    RegisterAnalyzerTestCase(
        X64_ARCH,
        Instruction(
            virtual_address=0x400491,
            size=1,
            disassembly="pop rbx",
            mnemonic="pop",
            operands="rbx",
            mode=InstructionSetMode.NONE,
        ),
        ("rsp",),
        ("rsp", "rbx"),
    ),
    RegisterAnalyzerTestCase(
        X64_ARCH,
        Instruction(
            virtual_address=0x400492,
            size=1,
            disassembly="leave ",
            mnemonic="leave",
            operands="",
            mode=InstructionSetMode.NONE,
        ),
        ("rbp", "rsp"),
        ("rbp", "rsp"),
    ),
    RegisterAnalyzerTestCase(
        X64_ARCH,
        Instruction(
            virtual_address=0x400493,
            size=1,
            disassembly="ret ",
            mnemonic="ret",
            operands="",
            mode=InstructionSetMode.NONE,
        ),
        ("rsp",),
        ("rsp",),
    ),
]

PPC_REGISTER_USAGE_TEST_CASES = [
    RegisterAnalyzerTestCase(
        PPC_ARCH,
        Instruction(
            0x100,
            0x4,
            "add r4, r5, r6",
            "add",
            "r4, r5, r6",
            InstructionSetMode.NONE,
        ),
        ("r4", "r5", "r6"),
        ("r4",),
    ),
]


REGISTER_USAGE_TEST_CASES = (
    ARM_REGISTER_USAGE_TEST_CASES + X64_REGISTER_USAGE_TEST_CASES + PPC_REGISTER_USAGE_TEST_CASES
)


class RegisterUsageTestPattern:
    @pytest.fixture
    async def assembler_service(self, ofrak_context: OFRAKContext):
        return await ofrak_context.injector.get_instance(AssemblerServiceInterface)

    def case_is_known_broken(self, test_case: RegisterAnalyzerTestCase) -> Tuple[bool, str]:
        """
        Reimplement this method with a check to find test cases we know should break for an
        implementation.

        This method can return a string with the reason the test case is known to be broken. The
        returned string will be displayed as the reason the test was skipped.

        :return: A tuple of True or False for is or is not known broken, and a string with the
        reason the test case is broken.
        """
        return False, ""

    @pytest.mark.parametrize("test_case", REGISTER_USAGE_TEST_CASES, ids=lambda tc: tc.label)
    async def test_register_usage_analyzer(
        self, test_case: RegisterAnalyzerTestCase, ofrak_context, assembler_service
    ):
        test_case_is_broken, reason = self.case_is_known_broken(test_case)
        if test_case_is_broken:
            if not reason:
                reason = "test case is known to be broken"
            pytest.skip(reason)

        await self._get_and_check_register_usage(test_case, ofrak_context, assembler_service)

    @pytest.mark.parametrize("test_case", REGISTER_USAGE_TEST_CASES, ids=lambda tc: tc.label)
    async def test_known_broken_cases(
        self, test_case: RegisterAnalyzerTestCase, ofrak_context, assembler_service
    ):
        """
        Test that cases which are marked as known broken are still broken. If something has changed
        to make them no longer break as expected, the test must be updated.
        """
        test_case_is_broken, _ = self.case_is_known_broken(test_case)
        if not test_case_is_broken:
            pytest.skip("test case not broken")

        with pytest.raises(Exception):
            await self._get_and_check_register_usage(test_case, ofrak_context, assembler_service)

    async def _get_and_check_register_usage(
        self, test_case: RegisterAnalyzerTestCase, ofrak_context, assembler_service
    ):
        instr_data = await assembler_service.assemble(
            test_case.instruction.disassembly,
            test_case.instruction.virtual_address,
            test_case.program_attributes,
            test_case.instruction.mode,
        )
        instr_r = await ofrak_context.create_root_resource(
            "test_resource", instr_data, tags=(Instruction,)
        )
        instr_r.add_view(test_case.instruction)
        instr_r.add_attributes(test_case.program_attributes)
        await instr_r.save()

        register_usage = await instr_r.analyze(RegisterUsage)

        assert set(test_case.expected_regs_read) == set(
            register_usage.registers_read
        ), self._pretty_print_fail(test_case.expected_regs_read, register_usage.registers_read)
        assert set(test_case.expected_regs_written) == set(
            register_usage.registers_written
        ), self._pretty_print_fail(
            test_case.expected_regs_written, register_usage.registers_written
        )

    @staticmethod
    def _pretty_print_fail(expected, actual):
        return f"missing: {set(expected).difference(actual)}, unexpected: {set(actual).difference(expected)}"



import dataclasses
from dataclasses import is_dataclass, fields
from typing import Any, Dict, Type, cast, Tuple

import inspect

from ofrak.service.serialization.pjson_types import PJSONType
from ofrak.service.serialization.serializers.enum_serializer import is_enum
from ofrak.service.serialization.serializers.serializer_i import SerializerInterface
from ofrak.service.serialization.serializers.type_serializer import is_metaclass


def _is_regular_class_instance(type_hint):
    return inspect.isclass(type_hint) and not is_enum(type_hint) and not is_metaclass(type_hint)


class ClassInstanceSerializer(SerializerInterface):
    """
    Serialize and deserialize class instances into `PJSONType`.

    Implementation: the class is serialized, along with the attributes of the instance taken from its annotations.
    In other words, the annotations determine the exact attributes of the instance which are serialized and then
    re-created during deserialization.

    Annotations are present in dataclasses by construction, but they need to be added for all the other classes
    if this default serializer is to process them correctly.

    Invoking __init__ and __post_init__ during deserialization could have side effects, though.

    For dataclasses, __init__ corresponds perfectly to the annotations, and the default assumption is that
    side effects in __post_init__ are desirable to re-invoke during deserialization (if not, write a custom
    serializer).
    For other classes, the default assumption is that __init__ and __post_init__ might not correspond to the
    annotations and that their side effects are undesirable for deserialization, so we only call __new__
    and then manually `setattr` the fields. Write a custom serializer if this behavior doesn't work well
    for some class. For example, the following class will be serialized and deserialized correctly, but wouldn't
    if __init__ was called:

    ``` py
    class Example:
        i: int
        def __init__(self, i: int):
            self.i = i + 1
    ```

    The type hint (name of the class) given is ignored, instead we get the annotations from the class instance
    itself. This allows to deal with the frequent case of a type hint corresponding to a superclass of the class
    instance, e.g. an interface, which wouldn't provide the correct annotations and would require a custom
    serializer for the superclass (which would dispatch based on the actual type of the class instance given).

    The more precise class is used in the serialized form, allowing deserialization to know which precise class
    to handle as well.
    """

    targets = (_is_regular_class_instance,)

    def obj_to_pjson(self, cls_instance: Any, _type_hint: Any) -> Tuple[str, Dict[str, PJSONType]]:
        cls_fields_pjson: Dict[str, PJSONType] = {}
        fields_and_types = self._get_class_fields_and_types(
            cls_instance, as_dataclass=is_dataclass(cls_instance)
        )
        for field_name, field_type in fields_and_types.items():
            cls_fields_pjson[field_name] = self._service.to_pjson(
                getattr(cls_instance, field_name), field_type
            )
        cls_ref_pjson = self._service.to_pjson(cls_instance.__class__, Type)
        cls_ref_pjson = cast(str, cls_ref_pjson)
        return (cls_ref_pjson, cls_fields_pjson)

    def pjson_to_obj(self, pjson_obj: Tuple[str, Dict[str, PJSONType]], _type_hint: Any) -> Any:
        cls_ref_pjson, cls_fields_pjson = pjson_obj
        cls = self._service.from_pjson(cls_ref_pjson, Type)
        return self._deserialize_instance(cls, cls_fields_pjson)

    @staticmethod
    def _get_class_fields_and_types(cls: Any, as_dataclass: bool) -> Dict[str, Any]:
        """Return the field names and types for `cls`, a class or class instance."""
        if as_dataclass is True:
            # Fields for which init is False aren't serialized
            return {field.name: field.type for field in fields(cls) if field.init is True}
        else:
            # NOTE: in python >= 3.10, inspect.get_annotations() is a better choice
            return getattr(cls, "__annotations__", {})

    def _deserialize_instance(self, cls: Any, cls_fields_pjson: Dict[str, PJSONType]) -> Any:
        expected_fields_and_types = self._get_class_fields_and_types(
            cls, as_dataclass=is_dataclass(cls)
        )
        deserialized_fields = {
            field_name: self._service.from_pjson(cls_fields_pjson[field_name], field_type)
            for field_name, field_type in expected_fields_and_types.items()
        }
        if is_dataclass(cls) and getattr(cls, dataclasses._PARAMS).init:  # type: ignore
            return cls(**deserialized_fields)
        else:
            cls_instance = cls.__new__(cls)
            for field_name, field in deserialized_fields.items():
                object.__setattr__(cls_instance, field_name, field)
            return cls_instance



import os
from dataclasses import dataclass
from typing import Dict, List, Union

import pytest
from ofrak.core.filesystem import File

from ofrak import OFRAKContext
from ofrak_type.architecture import InstructionSetMode
from ofrak.core.basic_block import BasicBlock
from ofrak.core.complex_block import ComplexBlock
from ofrak.core.data import DataWord
from ofrak.resource import Resource
from ofrak.service.resource_service_i import ResourceFilter, ResourceSort
from ofrak.core.elf.model import Elf
from pytest_ofrak.patterns import TEST_PATTERN_ASSETS_DIR
from pytest_ofrak.patterns.unpack_verify import (
    UnpackAndVerifyTestCase,
    UnpackAndVerifyPattern,
)


@dataclass
class ComplexBlockUnpackerTestCase(UnpackAndVerifyTestCase[int, List[Union[BasicBlock, DataWord]]]):
    binary_filename: str
    binary_md5_digest: str


COMPLEX_BLOCK_UNPACKER_TEST_CASES = [
    ComplexBlockUnpackerTestCase(
        "x64",
        {
            0x4003E0: [
                BasicBlock(
                    virtual_address=0x4003E0,
                    size=42,
                    mode=InstructionSetMode.NONE,
                    is_exit_point=True,
                    exit_vaddr=None,
                ),
            ],
            0x40040C: [
                BasicBlock(
                    virtual_address=0x40040C,
                    size=16,
                    mode=InstructionSetMode.NONE,
                    is_exit_point=False,
                    exit_vaddr=4195356,
                ),
                BasicBlock(
                    virtual_address=0x40041E,
                    size=5,
                    mode=InstructionSetMode.NONE,
                    is_exit_point=True,
                    exit_vaddr=None,
                ),
                BasicBlock(
                    virtual_address=0x40041C,
                    size=2,
                    mode=InstructionSetMode.NONE,
                    is_exit_point=False,
                    exit_vaddr=4195358,
                ),
            ],
            0x400430: [
                BasicBlock(
                    virtual_address=0x400430,
                    size=18,
                    mode=InstructionSetMode.NONE,
                    is_exit_point=False,
                    exit_vaddr=4195394,
                ),
                BasicBlock(
                    virtual_address=0x40048D,
                    size=7,
                    mode=InstructionSetMode.NONE,
                    is_exit_point=True,
                    exit_vaddr=None,
                ),
                BasicBlock(
                    virtual_address=0x400442,
                    size=32,
                    mode=InstructionSetMode.NONE,
                    is_exit_point=False,
                    exit_vaddr=4195426,
                ),
                BasicBlock(
                    virtual_address=0x400486,
                    size=7,
                    mode=InstructionSetMode.NONE,
                    is_exit_point=False,
                    exit_vaddr=4195469,
                ),
                BasicBlock(
                    virtual_address=0x400462,
                    size=6,
                    mode=InstructionSetMode.NONE,
                    is_exit_point=False,
                    exit_vaddr=4195432,
                ),
                BasicBlock(
                    virtual_address=0x400468,
                    size=30,
                    mode=InstructionSetMode.NONE,
                    is_exit_point=False,
                    exit_vaddr=4195462,
                ),
            ],
            0x4004A0: [
                BasicBlock(
                    virtual_address=0x4004A0,
                    size=14,
                    mode=InstructionSetMode.NONE,
                    is_exit_point=False,
                    exit_vaddr=4195502,
                ),
                BasicBlock(
                    virtual_address=0x4004C0,
                    size=2,
                    mode=InstructionSetMode.NONE,
                    is_exit_point=True,
                    exit_vaddr=None,
                ),
                BasicBlock(
                    virtual_address=0x4004AE,
                    size=10,
                    mode=InstructionSetMode.NONE,
                    is_exit_point=False,
                    exit_vaddr=4195512,
                ),
                BasicBlock(
                    virtual_address=0x4004B8,
                    size=8,
                    mode=InstructionSetMode.NONE,
                    is_exit_point=True,
                    exit_vaddr=None,
                ),
            ],
            0x4004C4: [
                BasicBlock(
                    virtual_address=0x4004C4,
                    size=25,
                    mode=InstructionSetMode.NONE,
                    is_exit_point=True,
                    exit_vaddr=None,
                ),
            ],
            0x4004E0: [
                BasicBlock(
                    virtual_address=0x4004E0,
                    size=2,
                    mode=InstructionSetMode.NONE,
                    is_exit_point=True,
                    exit_vaddr=None,
                ),
            ],
            0x4004F0: [
                BasicBlock(
                    virtual_address=0x4004F0,
                    size=74,
                    mode=InstructionSetMode.NONE,
                    is_exit_point=False,
                    exit_vaddr=4195642,
                ),
                BasicBlock(
                    virtual_address=0x400556,
                    size=35,
                    mode=InstructionSetMode.NONE,
                    is_exit_point=True,
                    exit_vaddr=None,
                ),
                BasicBlock(
                    virtual_address=0x40053A,
                    size=6,
                    mode=InstructionSetMode.NONE,
                    is_exit_point=False,
                    exit_vaddr=4195648,
                ),
                BasicBlock(
                    virtual_address=0x400540,
                    size=22,
                    mode=InstructionSetMode.NONE,
                    is_exit_point=False,
                    exit_vaddr=4195670,
                ),
            ],
            0x400580: [
                BasicBlock(
                    virtual_address=0x400580,
                    size=22,
                    mode=InstructionSetMode.NONE,
                    is_exit_point=False,
                    exit_vaddr=4195734,
                ),
                BasicBlock(
                    virtual_address=0x4005AF,
                    size=7,
                    mode=InstructionSetMode.NONE,
                    is_exit_point=True,
                    exit_vaddr=None,
                ),
                BasicBlock(
                    virtual_address=0x400596,
                    size=10,
                    mode=InstructionSetMode.NONE,
                    is_exit_point=False,
                    exit_vaddr=4195744,
                ),
                BasicBlock(
                    virtual_address=0x4005A0,
                    size=15,
                    mode=InstructionSetMode.NONE,
                    is_exit_point=False,
                    exit_vaddr=4195759,
                ),
            ],
        },
        {
            # No optional results
        },
        "hello.out",
        "cc2de3c0cd2d0ded7543682c2470fcf0",
    ),
    # ARM with literal pools
    ComplexBlockUnpackerTestCase(
        "ARM with literal pools",
        {
            0x8018: [
                BasicBlock(
                    virtual_address=0x8018,
                    size=24,
                    mode=InstructionSetMode.NONE,
                    is_exit_point=True,
                    exit_vaddr=None,
                ),
                DataWord(virtual_address=32816, size=4, format_string="<L", xrefs_to=(32792,)),
            ],
            0x8034: [
                BasicBlock(
                    virtual_address=0x8034,
                    size=20,
                    mode=InstructionSetMode.NONE,
                    is_exit_point=False,
                    exit_vaddr=32840,
                ),
                BasicBlock(
                    virtual_address=0x8058,
                    size=8,
                    mode=InstructionSetMode.NONE,
                    is_exit_point=True,
                    exit_vaddr=None,
                ),
                BasicBlock(
                    virtual_address=0x8048,
                    size=16,
                    mode=InstructionSetMode.NONE,
                    is_exit_point=False,
                    exit_vaddr=32856,
                ),
                DataWord(virtual_address=32864, size=4, format_string="<L", xrefs_to=(32820,)),
                DataWord(virtual_address=32868, size=4, format_string="<L", xrefs_to=(32840,)),
            ],
            0x8068: [
                BasicBlock(
                    virtual_address=0x8068,
                    size=40,
                    mode=InstructionSetMode.NONE,
                    is_exit_point=True,
                    exit_vaddr=None,
                ),
            ],
        },
        {
            # No optional results
        },
        "simple_arm_gcc.o.elf",
        "c79d1bea0398d7a9d0faa1ba68786f5e",
    ),
]


class ComplexBlockUnpackerUnpackAndVerifyPattern(UnpackAndVerifyPattern):
    """
    Test pattern which checks a ComplexBlockUnpacker implementation. This pattern is ready to go
    off-the-shelf. All that is needed to use it is
     1) A subclass with the name prefixed with "Test", and
     2) That subclass should be in a pytest context where the supplied frak context will have a
     ComplexBlockUnpacker implementation.

    This file includes test cases of the type ComplexBlockUnpackerTestCase. These include a URL
    which points to a binary which serves as the root resource. This binary will be unpacked
    recursively down to BasicBlocks, and all ComplexBlocks in the .text section will be
    extracted. Those ComplexBlocks should line up with the expected complex block virtual
    addresses; so should all of the BasicBlocks and DataWords that make up that ComplexBlock.
    Each of those are checked for some other specific expected attributes as well.
    """

    @pytest.fixture(params=COMPLEX_BLOCK_UNPACKER_TEST_CASES, ids=lambda tc: tc.label)
    async def unpack_verify_test_case(self, request) -> ComplexBlockUnpackerTestCase:
        return request.param

    @pytest.fixture
    async def root_resource(
        self,
        unpack_verify_test_case: ComplexBlockUnpackerTestCase,
        ofrak_context: OFRAKContext,
        test_id: str,
    ) -> Resource:
        asset_path = os.path.join(TEST_PATTERN_ASSETS_DIR, unpack_verify_test_case.binary_filename)
        with open(asset_path, "rb") as f:
            binary_data = f.read()
        resource = await ofrak_context.create_root_resource(test_id, binary_data, tags=(File,))
        return resource

    async def unpack(self, root_resource: Resource):
        await root_resource.unpack_recursively(do_not_unpack=(BasicBlock,))

    async def get_descendants_to_verify(self, unpacked_resource: Resource) -> Dict[int, Resource]:
        elf = await unpacked_resource.view_as(Elf)
        text_section = await elf.get_section_by_name(".text")
        complex_blocks: List[ComplexBlock] = list(
            await text_section.resource.get_descendants_as_view(
                ComplexBlock,
                r_filter=ResourceFilter.with_tags(ComplexBlock),
                r_sort=ResourceSort(ComplexBlock.VirtualAddress),
            )
        )
        return {cb.virtual_address: cb for cb in complex_blocks}

    async def verify_descendant(
        self, complex_block: ComplexBlock, specified_result: List[Union[BasicBlock, DataWord]]
    ):
        basic_blocks = await complex_block.get_basic_blocks()

        # Check that the parent complex blocks are extracted as expected
        complex_block_start_address = complex_block.virtual_address
        # Check that all expected basic blocks have been extracted
        assert {basic_block.virtual_address for basic_block in basic_blocks} == {
            expected_basic_block.virtual_address
            for expected_basic_block in specified_result
            if isinstance(expected_basic_block, BasicBlock)
        }

        blocks_by_addr: Dict[int, Union[DataWord, BasicBlock]] = {
            block.virtual_address: block for block in specified_result
        }
        for basic_block in basic_blocks:
            expected_basic_block = blocks_by_addr[basic_block.virtual_address]
            assert type(expected_basic_block) is BasicBlock, (
                f"got BasicBlock at "
                f"{hex(basic_block.virtual_address)} but expected {type(expected_basic_block)}"
            )
            expected_bb_info = (
                expected_basic_block.virtual_address,
                expected_basic_block.size,
                expected_basic_block.mode,
                expected_basic_block.is_exit_point,
                expected_basic_block.exit_vaddr,
            )
            extracted_bb_info = (
                basic_block.virtual_address,
                basic_block.size,
                basic_block.mode,
                basic_block.is_exit_point,
                basic_block.exit_vaddr,
            )
            assert expected_bb_info == extracted_bb_info, (
                f"Extracted BasicBlocks do not match expected for ComplexBlock "
                f"0x{complex_block_start_address:x}: got {extracted_bb_info}, expected {expected_bb_info}"
            )

        data_words = await complex_block.get_data_words()
        for data_word in data_words:
            expected_data_word = blocks_by_addr[data_word.virtual_address]
            assert type(expected_data_word) is DataWord, (
                f"got DataWord at "
                f"{hex(data_word.virtual_address)} but expected {type(expected_data_word)}"
            )
            expected_data_word_info = (
                expected_data_word.virtual_address,
                expected_data_word.size,
                expected_data_word.format_string,
                expected_data_word.xrefs_to,
            )
            extracted_data_word_info = (
                data_word.virtual_address,
                data_word.size,
                data_word.format_string,
                data_word.xrefs_to,
            )
            assert expected_data_word_info == extracted_data_word_info, (
                f"Extracted DataWord do not match expected for ComplexBlock: "
                f"0x{complex_block_start_address:x}"
            )



import warnings
from abc import abstractmethod, ABC
from dataclasses import dataclass
from typing import Any, Dict, Generic, TypeVar, Set, Iterable

import pytest

from ofrak import OFRAKContext
from ofrak.resource import Resource

K = TypeVar("K")
V = TypeVar("V")


def _hexlify(x):
    if isinstance(x, Iterable):
        return [hex(y) if isinstance(y, int) else y for y in x]
    return hex(x) if isinstance(x, int) else x


@dataclass
class UnpackAndVerifyTestCase(Generic[K, V]):
    """
    Container for test cases which will be fed to an UnpackAndVerifyPattern implementation.
    Extend this class, adding some fields to hold the information needed to create the root
    resource for each test case.

    Test cases are organized into "expected" and "optional" results.
      - Expected results are required to exist in the unpacker output following the test
        procedure. The lack of any expected result fails the test.
      - Optional results may or may not exist in the unpacker output. The lack of an optional
        result generates a warning.
      - Results which aren't specified in either 'expected' or 'optional' sets fail the test.

    :param K: The key type of the expected and optional results. The unpacked resources will each be assigned
    a key of this same type, allowing the test to match actual unpacked resources with expected
    values.
    :param V: The value type of the expected and optional results. This can be anything, as long as it
    contains the information needed to verify that the actual unpacked resource matches what is
    expected.
    """

    label: str
    expected_results: Dict[K, V]
    optional_results: Dict[K, V]


class UnpackAndVerifyPattern(ABC):
    """
    Test pattern for unpackers. The test expects to:
    1. Receive some parameterized test cases (fixture ``unpack_verify_test_case``)
    2. From each test case:
        a) Instantiate a root resource to unpack (fixture ``root_resource``)
        b) Extract the expected results (fixture ``expected_results``)
        c) Extract the optional results (fixture ``optional_results``)
    3. Unpack the root resource (method ``unpack``)
    4. Get the descendants of the unpacked root resource (method ``get_descendants_to_verify``)
    5. Test that the descendants match the expected and optional results, (method ``test_unpack_verify``), as follows:
        - Assert that all expected results exist as descendants of the unpacked resource
        - Assert that there are no descendants that are neither in the expected results nor in the optional results
        - Warn if any optional results do not exist as descendants of the unpacked resource

    Each step is broken out into a method. Some of the methods are abstract and need to be
    implemented in classes implementing this test pattern; others already implemented and only
    need to be re-implemented (overridden) in specific circumstances, explained in each method.
    """

    async def test_unpack_verify(
        self,
        root_resource: Resource,
        expected_results: Dict,
        optional_results: Dict,
    ):
        await self.unpack(root_resource)
        print(await root_resource.summarize_tree())

        ## Prepare results for comparison
        unpacked_results = await self.get_descendants_to_verify(root_resource)
        unpacked_set = set(unpacked_results.keys())

        expected_set = set(expected_results.keys())
        optional_set = set(optional_results.keys())

        unpacked_expected_set = unpacked_set & expected_set
        unpacked_optional_set = unpacked_set & optional_set

        # Optional entries are allowed to be absent from the unpacker results, so we warn
        # about them instead.
        missing_optional_set = optional_set - unpacked_set

        ## Build an info string about this test case
        info_str = [f"{'item':<16}{'unpacked':<16}{'expected':<16}{'optional':<16}"]
        for item in sorted(unpacked_set | expected_set | optional_set):
            item_fmt = str(_hexlify(item))
            row = (
                f"{item_fmt:<16}"
                f"{item_fmt if item in unpacked_set else '':<16}"
                f"{item_fmt if item in expected_set else '':<16}"
                f"{item_fmt if item in optional_set else '':<16}"
            )
            info_str.append(row)
        info_str = "\n".join(info_str)

        ## Sanity check to ensure that optional and expected sets are exclusive
        assert expected_set & optional_set == set(), (
            f"{_hexlify(expected_set & optional_set)} cannot be both expected and optional!\n"
            f"{info_str}"
        )

        ## Verify cardinality of expected descendants and unpacked descendants
        self.verify_expected_descendants(unpacked_expected_set, expected_set, info_str)

        ## Verify that there are no results outside the (expected | optional) sets
        self.verify_no_extraneous_descendants(unpacked_set, expected_set, optional_set, info_str)

        ## Verify the value of each expected descendant
        for key in expected_set:
            await self.verify_descendant(unpacked_results[key], expected_results[key])

        ## Verify the value of each unpacked optional descendant
        for key in unpacked_optional_set:
            await self.verify_descendant(unpacked_results[key], optional_results[key])

        ## Warn if there are missing optional results
        if missing_optional_set:
            warnings.warn(
                UserWarning(
                    f"The following optional functions are missing from the analyzer results: "
                    f"{_hexlify(sorted(missing_optional_set))}\n"
                    f"{info_str}"
                )
            )

    @pytest.fixture(params=[], ids=lambda tc: tc.label)
    @abstractmethod
    async def unpack_verify_test_case(self, request) -> UnpackAndVerifyTestCase:
        """
        Gathers all the test cases for a test class implementing this pattern

        Override this method to specify the parametrized test cases. The overriding method should
        look exactly the same as this base method, except with a list of actual test cases
        supplied to the ``params`` in the fixture decorator. The type of this list should be
        List[UnpackAndVerifyTestCase]
        """
        return request.param

    @pytest.fixture
    @abstractmethod
    async def root_resource(
        self,
        unpack_verify_test_case: UnpackAndVerifyTestCase,
        ofrak_context: OFRAKContext,
        test_id: str,
    ) -> Resource:
        """
        Instantiate a root resource for this test
        """
        raise NotImplementedError()

    @pytest.fixture
    async def expected_results(
        self, unpack_verify_test_case: UnpackAndVerifyTestCase[K, V]
    ) -> Dict[K, V]:
        """
        Extract the expected results from the test case.

        Override this method if for some reason the expected results need to be modified in some way
        before being passed to the test. If the expected results are good as-is, do not override
        this method.
        """
        return unpack_verify_test_case.expected_results

    @pytest.fixture
    async def optional_results(
        self, unpack_verify_test_case: UnpackAndVerifyTestCase[K, V]
    ) -> Dict[K, V]:
        """
        Extract the optional results from the test case.

        Override this method if for some reason the optional results need to be modified in some way
        before being passed to the test. If the optional results are good as-is, do not override
        this method.
        """
        return unpack_verify_test_case.optional_results

    @abstractmethod
    async def unpack(self, root_resource: Resource):
        """
        Unpack the root resource
        """
        raise NotImplementedError()

    @abstractmethod
    async def get_descendants_to_verify(self, unpacked_root_resource: Resource) -> Dict:
        """
        Once the root resource has been unpacked, get some set of descendants of the root
        resource to check against the expected results. A key should be calculated for each of
        these resources. This key should match the type K in the UnpackAndVerifyTestCase type
        this test uses, because it will be used to look up the expected result for each
        descendant in the test case.

        :returns: Dict[K, V] where K and V match the type parameters to the
        UnpackAndVerifyTestCase type for this test
        """
        raise NotImplementedError()

    @abstractmethod
    async def verify_descendant(self, unpacked_descendant: Any, specified_result: Any):
        """
        Verify that the actual unpacked descendant matches the specified result from the test case
        """
        raise NotImplementedError()

    def verify_expected_descendants(
        self, unpacked_set: Set, expected_set: Set, info_str: str = f""
    ):
        """
        Assert that the resources returned by ``get_descendants_to_verify`` (unpacked_set)
        match all values specified in ``expected_set``.

        Override to allow for unpacked results with missing expected values.
        """

        assert expected_set - unpacked_set == set(), (
            f"unpacked descendants {_hexlify(unpacked_set)} did not match expected {_hexlify(expected_set)}\n"
            f"missing results: {_hexlify(expected_set - unpacked_set)}\n"
            f"{info_str}"
        )

    def verify_no_extraneous_descendants(
        self,
        unpacked_set: Set,
        expected_set: Set,
        optional_set: Set,
        info_str: str = f"",
    ):
        """
        Assert that the resources returned by ``get_descendants_to_verify`` (unpacked_set)
        are all covered by an expected or optional test case.

        Override to allow unpacked values beyond the scope of (expected | optional) cases.
        """
        full_set = expected_set | optional_set
        unpacked_extraneous_set = unpacked_set - full_set

        assert unpacked_extraneous_set == set(), (
            f"Unpacked descendant(s) {_hexlify(unpacked_extraneous_set)} "
            f"outside the range of expected + optional descendants.\n"
            f"{info_str}"
        )



from typing import Any, Iterable, Tuple, Union

from intervaltree import IntervalTree

from ofrak.service.serialization.pjson_types import PJSONType
from ofrak.service.serialization.serializers.serializer_i import SerializerInterface

IntervalTreeTupleType = Union[Tuple[int, int], Tuple[int, int, Any]]


class IntervalTreeSerializer(SerializerInterface):
    """
    Serialize and deserialize third-party object `IntervalTree` into `PJSONType`.

    Implementation: an `IntervalTree` is a binary lookup tree of intervals, each interval being
    an `Interval` object representing a 3-tuple (begin, end, data).
    As `IntervalTree` provides a `from_tuples` constructor, and has an `all_intervals` attribute,
    we can simply serialize it as its list of intervals (themselves serialized as tuples).
    """

    targets = (IntervalTree,)

    def obj_to_pjson(self, obj: IntervalTree, _type_hint: Any) -> PJSONType:
        return self._service.to_pjson(
            [(interval.begin, interval.end, interval.data) for interval in obj.all_intervals],
            Iterable[IntervalTreeTupleType],
        )

    def pjson_to_obj(self, pjson_obj: PJSONType, type_hint: Any) -> IntervalTree:
        return IntervalTree.from_tuples(
            self._service.from_pjson(pjson_obj, Iterable[IntervalTreeTupleType])
        )
